DEBUG		= -g -O2
WARNINGS	= -W -Wimplicit -Wreturn-type -Wswitch -Wtrigraphs -Wformat\
		  -Wchar-subscripts -Wuninitialized -Wparentheses\
		  -Wtemplate-debugging
CC		= gcc -ansi
CCC		= gcc
INSTALL		= install
ECHO		= echo

CCFLAGS		= -D NAFS -D XLIB $(DEBUG)
CPPFLAGS	= $(WARNINGS)
LDFLAGS		=
BINDIR		= /usr/local/bin
MANDIR		= /usr/local/man

SRCS		= misc.cc memory.cc inout.cc\
		  mc6809.cc mc6809in.cc da6809.cc\
		  mc6821.cc wd1793.cc mc146818.cc\
		  pia1.cc pia2.cc e2floppy.cc nafs.cc e2video.cc \
		  resource.cc gui.cc cpuview.cc main.cc
OBJS		= $(SRCS:.cc=.o)

BIN		= flexemu
LIBS		= -L/usr/X11/lib -lX11

$(BIN):		$(OBJS)
	$(CCC) -o $(@) $(CCFLAGS) $(LDFLAGS) $(OBJS) $(LIBS)

fromflex:	fromflex.cc
	$(CCC) -o $(@) $(CPPFLAGS) $(DEBUG) $(LDFLAGS) fromflex.cc

toflex:		toflex.cc
	$(CCC) -o $(@) $(CPPFLAGS) $(DEBUG) $(LDFLAGS) toflex.cc

.SUFFIXES:	.cc

.cc.o:
	$(CCC) $(CPPFLAGS) $(CCFLAGS) -c $<

$(OBJS):	machdep.h


machdep:	machdep.o
	$(CC) -o $(@) $(CPPFLAGS) $(CCFLAGS) $(LDFLAGS) machdep.o

machdep.h:	machdep
	./machdep $(@)

all: fromflex toflex $(BIN)
	@$(ECHO) "all binaries are up to date"
	@$(ECHO) "Now You may execute 'make install'"

install.bin:	fromflex toflex $(BIN)
	$(INSTALL) -m 755 -s fromflex toflex $(BIN) $(BINDIR)
	$(INSTALL) -m 644 test/neumon54.hex $(BINDIR)

install.man:
	$(INSTALL) -m 644 ../man/man1/flexemu.1 $(MANDIR)/man1
	$(INSTALL) -m 644 ../man/man2/flexdos.2 ../man/man2/flexfms.2\
		$(MANDIR)/man2
	$(INSTALL) -m 644  ../man/man7/flexerr.7 ../man/man7/flexfcb.7\
		../man/man7/flexmem.7 ../man/man7/flexuser.7\
		$(MANDIR)/man7

install.user:
	cp test/neumon54.hex $HOME/neumon54.hex

install:	install.bin install.man

uninstall:
	rm -f $(BINDIR)/fromflex $(BINDIR)/toflex \
		$(BINDIR)/$(BIN)
	rm -f $(MANDIR)/man1/flexemu.1 \
		$(MANDIR)/man2/flexdos.2 $(MANDIR)/man2/flexfms.2 \
		$(MANDIR)/man7/flexerr.7 $(MANDIR)/man7/flexfcb.7 \
		$(MANDIR)/man7/flexmem.7 $(MANDIR)/man7/flexuser.7

clean:
	$(RM) -f machdep.h machdep.o machdep\
	fromflex.o fromflex\
	toflex.o toflex\
	$(BIN) $(OBJS)

depend:		machdep.h
	makedepend $(SRCS)

# DO NOT DELETE THIS LINE -- make depend depends on it.
