/*
    cpuview.cc


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "gui.h"
#include "e2video.h"

#ifndef XLIB
int radio_data[] = {NO_CHANGE, RUN, STOP, STEP, EXIT, RESET, NEXT};

void Gui::popup_cpu(void)
{
	XtPopup(cpuframe, XtGrabNone);
}

void Gui::popdown_cpu(void)
{
	XtPopdown(cpuframe);
}

void Gui::popup_bp(unsigned int which)
{
	String bpstring;

	which_bp = which;
	bpstring = (String)XtMalloc(6);
	if (!cpu->is_bp_set(which))
		strcpy(bpstring, "");
	else
		sprintf(bpstring, "%04x", cpu->get_bp(which));
	XtVaSetValues(bptext, XtNlength, 6,
		XtNstring, (XtArgVal)bpstring, NULL);
	XtPopup(bpframe, XtGrabExclusive);
	XtFree(bpstring);
}

void Gui::popdown_bp(void)
{
	String bpstring;
	unsigned int addr, res;

	XtVaGetValues(bptext, XtNstring, &bpstring, NULL);
	if (sscanf(bpstring, "%x", (int *)&addr) == 1 && addr <= 0xffff)
		cpu->set_bp(which_bp, (Word)addr);
	else
		if (strlen(bpstring) == 0)
		 	cpu->reset_bp(which_bp);
	XtPopdown(bpframe);
	update_1_second(NO_CHANGE);
}

void Gui::clear_bp(void)
{
	XtVaSetValues(bptext, XtNstring, (XtArgVal)"", NULL);
}
#endif

#ifdef XLIB
void Gui::process_button_input(XEvent *pevent)
{
	unsigned int new_state = NO_CHANGE;
	int x, y, i, j;

	if (pevent->xbutton.button == Button1) {
	   // calculate character coordinates
	   x = (pevent->xbutton.x - 4) / cw;
	   y = (pevent->xbutton.y + fs->descent - 4) / ch;
	   if (y == BUTTON_RUN_Y && x >= BUTTON_X && x <= (BUTTON_X + 2))
		  new_state = RUN;
	   if (y == BUTTON_STOP_Y && x >= BUTTON_X && x <= (BUTTON_X + 3))
		new_state = STOP;
	   if (y == BUTTON_STEP_Y && x >= BUTTON_X && x <= (BUTTON_X + 3))
		new_state = STEP;
	   if (y == BUTTON_NEXT_Y && x >= BUTTON_X && x <= (BUTTON_X + 3))
		new_state = NEXT;
	   if (y == BUTTON_RESET_Y && x >= BUTTON_X && x <= (BUTTON_X + 4))
		new_state = RESET;
	   if (y == BUTTON_EXIT_Y && x >= BUTTON_X && x <= (BUTTON_X + 3))
		new_state = EXIT;
	   set_new_state(new_state);
	   for (i=0; i < 2; i++) {
		if (y == (5+i) && x >= 25 && x <= 28) {
			if (!bp_input[i]) {
				for (j=0; j < 2; j++)
					bp_input[j] = 0;
				bp_input[i] = 1;
				cpu->reset_bp(i);
			} else
				bp_input[i] = 0;
			break;
		}
	   }  // for
	   update_1_second(new_state);
	} // if
} // process_button_input
#endif

#ifndef XLIB
void Gui::create_cpu_view(Widget parent)
{
	cpuframe = XtVaCreatePopupShell("MC6809",
		topLevelShellWidgetClass, parent, NULL);
	cpuform = XtVaCreateManagedWidget("cpuForm",
		formWidgetClass, cpuframe, NULL);
	cputext = XtVaCreateManagedWidget("cpuText",
		asciiTextWidgetClass, cpuform, NULL);
	runbutton = XtVaCreateManagedWidget("runButton",
		toggleWidgetClass, cpuform,
		XtNradioData, (XtArgVal)&radio_data[RUN],
		XtNstate, (XtArgVal)1, NULL);
	stopbutton = XtVaCreateManagedWidget("stopButton",
		toggleWidgetClass, cpuform,
		XtNradioGroup, (XtArgVal)runbutton,
		XtNradioData, (XtArgVal)&radio_data[STOP], NULL);
	stepbutton = XtVaCreateManagedWidget("stepButton",
		toggleWidgetClass, cpuform,
		XtNradioGroup, (XtArgVal)runbutton,
		XtNradioData, (XtArgVal)&radio_data[STEP], NULL);
	nextbutton = XtVaCreateManagedWidget("nextButton",
		toggleWidgetClass, cpuform,
		XtNradioGroup, (XtArgVal)runbutton,
		XtNradioData, (XtArgVal)&radio_data[NEXT], NULL);
	resetbutton = XtVaCreateManagedWidget("resetButton",
		toggleWidgetClass, cpuform,
		XtNradioGroup, (XtArgVal)runbutton,
		XtNradioData, (XtArgVal)&radio_data[RESET], NULL);
	bp1button = XtVaCreateManagedWidget("bp1Button",
		commandWidgetClass, cpuform, NULL);
	bp2button = XtVaCreateManagedWidget("bp2Button",
		commandWidgetClass, cpuform, NULL);
	cpubutton = XtVaCreateManagedWidget("cpuButton",
		commandWidgetClass, cpuform, NULL);

	XtAddCallback(runbutton,   XtNcallback, buttonCallback, &button_id[3]);
	XtAddCallback(stopbutton,  XtNcallback, buttonCallback, &button_id[4]);
	XtAddCallback(stepbutton,  XtNcallback, buttonCallback, &button_id[5]);
	XtAddCallback(nextbutton,  XtNcallback, buttonCallback, &button_id[6]);
	XtAddCallback(resetbutton, XtNcallback, buttonCallback, &button_id[7]);
	XtAddCallback(bp1button,   XtNcallback, buttonCallback, &button_id[8]);
	XtAddCallback(bp2button,   XtNcallback, buttonCallback, &button_id[9]);
	XtAddCallback(cpubutton,   XtNcallback, buttonCallback, &button_id[10]);
} // create_cpu_view

void Gui::create_bp_dialog(Widget parent)
{
	bpframe = XtVaCreatePopupShell("breakpoint",
		transientShellWidgetClass, parent, NULL);
	bpform = XtVaCreateManagedWidget("bpForm",
		formWidgetClass, bpframe, NULL);
	bptext = XtVaCreateManagedWidget("bpText",
		asciiTextWidgetClass, bpform,
		XtNeditType, (XtArgVal)XawtextEdit, NULL);
	bpbutton = XtVaCreateManagedWidget("bpButton",
		commandWidgetClass, bpform, NULL);
	clearbutton = XtVaCreateManagedWidget("clearButton",
		commandWidgetClass, bpform, NULL);

	XtAddCallback(bpbutton,    XtNcallback, buttonCallback, &button_id[11]);
	XtAddCallback(clearbutton, XtNcallback, buttonCallback, &button_id[12]);

	XtInstallAllAccelerators(bpform, bptext);
} // create_bp_dialog
#endif

#ifdef XLIB
void Gui::initialize_cpu_window(void)
{
	char		      **ret;
	int			count;
	XSetWindowAttributes	attrs;
	XGCValues		gcv;
	XSizeHints		sh;
	XClassHint		xch;
	XEvent			event;

	// get a fixed font
	ret = XListFonts(display, "*-fixed-*-*-*-*-10-*", 1, &count);
	if (ret) {
		fs = XLoadQueryFont(display, ret[0]);
		if (!fs) {
			fs = XLoadQueryFont(display, "fixed");
		}
		XFreeFontNames(ret);
	} else {
		fs = XLoadQueryFont(display, "fixed");
	}
	cw = XTextWidth(fs, " ", 1);
	ch = (fs->ascent + fs->descent);


	// Setup default window attributes
	attrs.background_pixel		= BlackPixelOfScreen(screen);
	attrs.do_not_propagate_mask	= 0;
//	attrs.event_mask =	StructureNotifyMask |
	attrs.event_mask		= ButtonPressMask | ExposureMask |
					  KeyPressMask;

	// Create the window
	ww = cw * 38 + 8;
	wh = ch * 14 + 8;
	cpu_window = XCreateWindow(display, RootWindowOfScreen(screen),
		0, 0, ww, wh, 2, CopyFromParent, CopyFromParent, CopyFromParent,
		CWEventMask | CWBackPixel | CWDontPropagate, &attrs);

	// Create a pixmap
	pix = XCreatePixmap(display, RootWindowOfScreen(screen),
		ww, wh, DefaultDepthOfScreen(screen));

	// Create a GC to go with it
	gcv.foreground = WhitePixelOfScreen(screen);
	gcv.background = BlackPixelOfScreen(screen);
	gcv.font = fs->fid;
	cpu_gc = XCreateGC(display, pix, GCForeground | GCBackground | GCFont,
		&gcv);

	// And an inverted GC
	gcv.foreground = BlackPixelOfScreen(screen);
	gcv.background = WhitePixelOfScreen(screen);
	inv_gc = XCreateGC(display, pix, GCForeground | GCBackground | GCFont,
		&gcv);

	// Set window size
	sh.width = sh.max_width = ww;
	sh.height = sh.max_height = wh;
	sh.flags = PSize | PMaxSize;
	XSetStandardProperties(display, cpu_window, "MC6809", "mc6809",
		None, NULL, 0, &sh);

	// Set window class
	xch.res_name = "mc6809";
	xch.res_class = "MC6809";
	XSetClassHint(display, cpu_window, &xch);

	// Put the window on the screen
	XMapWindow(display, cpu_window);
	// wait until window is visible
	XWindowEvent(display, cpu_window, ExposureMask, &event);
	
	update_all(); // now display all
}
#endif


void Gui::update_all()
{
	int i;

#ifdef XLIB
	// clear the pixmap
	XFillRectangle(display, pix, inv_gc, 0, 0, ww, wh);
#else
	for (i = 1; i <= 14*LINE_SIZE; i++)
		cpustring[i-1] = (i % LINE_SIZE == 0) ? '\n' : ' ';
#endif
	text( 0, 0, "Cycl:");
	text( 0, 1, "Inst:");
	text( 0, 2, "  PC:");
	text( 0, 3, "   S:");
	text( 0, 4, "   U:");
	text( 0, 5, "   X:");
	text( 0, 6, "   Y:");

	text(15, 2, "EFHINZVC");
	text(10, 3, "  CC:");
	text(10, 4, "  DP:");
	text(10, 5, "   A:");
	text(10, 6, "   B:");
	text(19, 5, "  bp1:");
	text(19, 6, "  bp2:");
	text(20, 0, "Freq:");
	text(30, 0, "KHz");

	for (i = 0; i < 6; ++i) {
		text(4, i + 8, ":");
	}
	update_1_second(NO_CHANGE);
} // update_all


void Gui::text(int x, int y, const char *str, int rev)
{
#ifdef XLIB
	x = x * cw + 4;
	y = (y + 1) * ch + 4 - fs->descent;
	XDrawImageString(display, pix, rev ? inv_gc : cpu_gc, x, y, str,
		strlen(str));
#else
	strncpy(&cpustring[LINE_SIZE*y + x], str, strlen(str));
#endif
}

void Gui::update_1_second(unsigned int state_to_show)
{
	static char	tmp[40];
	unsigned int i, j;
	char *pmnem_buf, *pbuffer;
	int step;
	struct s_cpu_status stat;
	unsigned int state;

	cpu->get_status(&stat);
	text( 6, 2, hexstr(stat.pc));
	text( 6, 3, hexstr(stat.s));
	text( 6, 4, hexstr(stat.u));
	text( 6, 5, hexstr(stat.x));
	text( 6, 6, hexstr(stat.y));
	text(15, 3, binstr(stat.cc));
	text(15, 4, hexstr(stat.dp));
	text(15, 5, hexstr(stat.a));
	text(15, 6, hexstr(stat.b));
	sprintf(tmp, " %10lu", stat.cycles);
	text(6,  0, tmp);
	sprintf(tmp, "%4u", stat.freq > 9999 ? 9999 : stat.freq);
	text(25,  0, tmp);
	text(6, 1, "                        "); // first clear area
	if (cpu->disassemble((unsigned int)stat.pc, &step, &pbuffer, &pmnem_buf))
		text(6, 1, pmnem_buf);
	else
		text(6, 1, "sorry, no disassembler installed");
	if (state_to_show == NO_CHANGE)
		state = stat.state;
	else
		state = state_to_show;
#ifdef XLIB
	text(BUTTON_X, BUTTON_RUN_Y,   "RUN",   state == RUN);
	text(BUTTON_X, BUTTON_STOP_Y,  "STOP",  state == STOP);
	text(BUTTON_X, BUTTON_STEP_Y,  "STEP",  state == STEP);
	text(BUTTON_X, BUTTON_NEXT_Y,  "NEXT",  state == NEXT);
	text(BUTTON_X, BUTTON_RESET_Y, "RESET", state == RESET);
	text(BUTTON_X, BUTTON_EXIT_Y,  "EXIT",  state == EXIT);
#endif
	for (i=0; i < 2; i++) {
		if (!cpu->is_bp_set(i))
#ifdef XLIB
			text(25, 5+i, "    ", bp_input[i]);
#else
			text(25, 5+i, "    ", 0);
#endif
		else
			text(25, 5+i,
#ifdef XLIB
				hexstr((short unsigned int)cpu->get_bp(i)), bp_input[i]);
#else
				hexstr((short unsigned int)cpu->get_bp(i)), 0);
#endif
	}  // for

	Word		stk = ((stat.s >> 3) << 3) - 16;
	Word		stk_tmp = stk;

	for (i = 0; i < 6; ++i) {
		text(0, i + 8, hexstr(stk));

		tmp[0] = ' ';
		tmp[1] = '\0';
		for (j = 0; j < 8; ++j) {
			Byte		ch = memory->read(stk + j);
			strcat(tmp, hexstr(ch));
			strcat(tmp, " ");
		} // for

		for (j = 0; j < 8; ++j) {
			Byte		ch = memory->read(stk + j);
			strcat(tmp, ascchr(ch));
		} // for
		text(5, i + 8, tmp);
		stk+= 8;
	} // for
	i = stat.s & 7;
//	j = (stat.s & 0xf) / 8;
#ifdef XLIB
	text(6 + 3 * i, 10, hexstr(memory->read(stat.s)), 1);
	text(30 + i, 10, ascchr(memory->read(stat.s)), 1);
	XCopyArea(display, pix, cpu_window, cpu_gc, 0, 0, ww, wh, 0, 0);
#else
	text(5 + 3 * i, 10, "[", 0);
	text(8 + 3 * i, 10, "]", 0);
	XtVaSetValues(cputext, XtNlength, strlen(cpustring)+1,
			XtNstring, cpustring, NULL);
	XawToggleSetCurrent(runbutton, (XtPointer)&radio_data[state]);
	if (state == RUN) {
		XtVaSetValues(stepbutton, XtNsensitive, 0, NULL);
		XtVaSetValues(nextbutton, XtNsensitive, 0, NULL);
		XtVaSetValues(resetbutton, XtNsensitive, 0, NULL);
	} else {
		XtVaSetValues(stepbutton, XtNsensitive, 1, NULL);
		XtVaSetValues(nextbutton, XtNsensitive, 1, NULL);
		XtVaSetValues(resetbutton, XtNsensitive, 1, NULL);
	}
#endif
}

#ifdef XLIB
void Gui::process_bp_input(KeySym keysym, int which)
{
	unsigned int digit;
	char ch;

	switch (keysym) {
		case XK_0 :
		case XK_1 :
		case XK_2 :
		case XK_3 :
		case XK_4 :
		case XK_5 :
		case XK_6 :
		case XK_7 :
		case XK_8 :
		case XK_9 :
		case XK_a :
		case XK_b :
		case XK_c :
		case XK_d :
		case XK_e :
		case XK_f :
			ch = (char)keysym;
			digit = isdigit(ch) ? ch - '0' : ch - 'a' + 10;
			if (!cpu->is_bp_set(which))
				cpu->set_bp(which, 0);
			if (cpu->get_bp(which) < 0x1000)
				cpu->set_bp(which, 16 * cpu->get_bp(which) + digit);
			break;
		case XK_Delete: cpu->reset_bp(which);
		case XK_Return: bp_input[which] = 0;
			break;
	} // switch
}
#endif

