/*
    e2floppy.h


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/


#ifndef __e2floppy_h__
#define __e2floppy_h__

#ifndef __fromflex__
#include <stdio.h>
#ifdef ultrix
#include <sys/types.h>
#endif // #ifdef ultrix
#include <sys/stat.h>
#include <unistd.h>
#include "mc6809.h"
#include "wd1793.h"

// possible floppy types:

#define NO_FORMAT  0	// unformated
#define OWN_FORMAT 1	// format defined for this emulator
#define BEN_FORMAT 2	// format used for Ben Williams FLEX emulator
#ifdef NAFS
#define NAFS_FORMAT 3	// map floppy to a UNIX directory (Unix File System)
#endif

#ifndef SEEK_SET
#define SEEK_SET 0
#endif /* #indef SEEK_SET */
#endif /* #ifndef __fromflex__ */

#define MAGIC_NUMBER 0x485c9a33L

/* datastructure describing the header of a FLEX disk image file */

struct s_flex_header {
	unsigned long magic_number;	/* to identify right format		 */
	unsigned char write_protect;	/* if != 0 disk image is write-protected */
	unsigned char sizecode;		/* 128 * 2 ^ n Bytes per Sector		 */
	unsigned char sides0;		/* nr. of sides on track 0		 */
	unsigned char sectors0;		/* nr. of sectors on track 0 (on one side) */
	unsigned char sides;		/* nr. of sides on track != 0		 */
	unsigned char sectors;		/* nr. of sectors on track != 0		 */
	unsigned char tracks;		/* nr. of tracks total			 */
	unsigned char dummy1;		/* for future extension			 */
	unsigned char dummy2, dummy3, dummy4, dummy5; /* and for alignment 4	 */
};

#ifndef __fromflex__

struct s_formats {
	long		size;		// byte size of disk
	unsigned int	tracks;		// number of tracks
	unsigned int	sectors;	// number of sectors
};

#ifdef NAFS

#define MAX_TRACK	79		// max. nr. of tracks - 1
#define MAX_SECTOR	36		// nr of sect. per track (side 0 and 1)
#define INIT_DIR_SECTORS (20-4)		// initial nr. of directory sectors
#define LINK_TABLE_SIZE ((MAX_TRACK+1) * MAX_SECTOR)
#define INIT_NEW_FILES	4		// initial nr of new files to be managed at a time
#define FREE_CHAIN	-1
#define DIRECTORY	-2
#define SYSTEM  	-3
#define NEW_FILE1	-4

struct s_st {
#ifdef MACH_BYTE_ORDER_MSB_FIRST
	unsigned char sec;
	unsigned char trk;
#else
	unsigned char trk;
	unsigned char sec;
#endif
};

typedef union {
	unsigned short sec_trk;		// spec. sector and tracknr. of as sect.
	struct s_st    st;
} t_st;

struct s_link_table {
	t_st		next;		// sector and tracknumber of next sect.
	unsigned char	record_nr[2];	// FLEX record number
	unsigned short	f_record;	// rel position in file / 252
	short		file_id;	// index of file in directory
};

// structure of one FLEX directory entry
struct s_dir_entry {
	char 		filename[8];
	char 		file_ext[3];
	unsigned char 	file_attr;
	unsigned char	reserved1;
	unsigned char 	start_trk;
	unsigned char 	start_sec;
	unsigned char   end_trk;
	unsigned char   end_sec;
	unsigned char 	records[2];
	unsigned char	sector_map;
	unsigned char	reserved2;
	unsigned char	month;
	unsigned char	day;
	unsigned char	year;
};

// structure of one FLEX directory sector
struct s_dir_sector {
	unsigned char next_trk;
	unsigned char next_sec;
	unsigned char record_nr[2];
	unsigned char unused[12];
	struct s_dir_entry dir_entry[10];
};

// structure of FLEX system info sector
struct s_sys_info_sector {
	unsigned char	unused1[16];
	char		disk_name[8];
	char		disk_ext[3];
	unsigned char	disk_number[2];
	unsigned char	fc_start_trk;
	unsigned char	fc_start_sec;
	unsigned char	fc_end_trk;
	unsigned char	fc_end_sec;
	unsigned char	free[2];
	unsigned char	month;
	unsigned char	day;
	unsigned char	year;
	unsigned char	last_trk;
	unsigned char	last_sec;
	unsigned char	unused2[216];
};

struct s_unused_sector {
	unsigned char	next_trk;
	unsigned char	next_sec;
	unsigned char	unused[254];
};

struct s_new_file {
	char filename[13];
	t_st first;
	t_st next;
	unsigned short f_record;
	FILE *fp;
};
#endif // #ifdef NAFS


struct s_floppy {
	unsigned char ready;
	FILE *fp;
	unsigned char  write_protect;
	unsigned int   offset;
	unsigned short byte_p_track;
	unsigned short byte_p_track0;
	unsigned short byte_p_sector;
	unsigned short max_sector;	// max. nr of sectors per side
	unsigned short max_sector0;	// max. nr of sectors per side track 0
	unsigned short max_track;	// max. tracknumber of disk 
	unsigned short track;		// actual tracknumber
	unsigned short type;		// type of floppy
	char 	      *ppath;		// pointer to unix path of drive
#ifdef NAFS
	s_link_table  *pflex_links;	// pointer to link table
	s_sys_info_sector *pflex_sys_info; // pointer to flex info sectors
	s_unused_sector *pflex_unused;  // pointer to flex unused sector (2)
	s_dir_sector  *pflex_directory; // pointer to flex directory
	s_new_file    *pnew_file;       // pointer to new file table
	int           dir_sectors;      // nr. of dir sectors in flex_directory
	int           new_files;        // nr. of new file entries 
	t_st          dir_extend;       //  track/sector of dir. extension sect.
#endif // #ifdef NAFS
};


class E2floppy : virtual public Wd1793 {

//
// internal registers
//
//	status			status register (r)
//	drisel			drive selection (w)
//	selected		selected drive as index into struct floppy

protected:

	Byte			status, drisel, selected;

// interface to operating system:
// a floppy structure for any drive connected
// drive nr. 4 means no drive selected
// pfs always points to the selectd floppy

protected:

	struct s_floppy		floppy[5], *pfs;
	char			sector_buffer[256];

// Internals

private:
	Mc6809*			cpu;
	virtual int		positionDrive(void);
protected:

	virtual Byte		 readByte(unsigned int index);
	virtual void		 writeByte(unsigned int index);
	virtual int		 driveReady(void);
	virtual int		 writeProtect(void);
	virtual int		 recordNotFound(void);
	virtual int		 seekError(Byte new_track);
	void			 header_for_own_format(
					s_floppy *floppy,
					s_flex_header *pheader,
					int wp);
	void			 header_for_ben_format(
					s_floppy *floppy,
					s_formats *pformat,
					int wp);
	char 			*open_mode(char *path);
	int			 init_disk_image(struct s_floppy *pfs);
	int			 init_nafs_drive(struct s_floppy *pfs);
	void			 mount_nafs(struct s_floppy *pfs,
					int index);

#ifdef NAFS
	void			header_for_nafs_format(
					s_floppy *floppy,
					int wp);
	void			fill_flex_directory(struct s_floppy *pfs,
					int dwp);
	void			initialize_flex_sys_info_sectors(
					struct s_floppy *pfs,
					int number);
	void			initialize_flex_unused_sector(struct s_floppy *pfs);
	void			initialize_flex_directory(struct s_floppy *pfs);
	void			initialize_flex_link_table(struct s_floppy *pfs);
	void			initialize_new_file_table(struct s_floppy *pfs);
	void			close_new_files(int index);
	void			free_nafs_memory(int index);
	void			umount_nafs(int index);
	int			open_files(int index);
	int			next_free_dir_entry(struct s_floppy *pfs);
	char			*unix_filename(int dir_index);
	char			*to_unix_filename(s_dir_entry *pde);
	void			scan_unix_filename(char *pfn, char *pfilename);
	int			add_to_link_table(
					struct s_floppy *pfs,
					int dir_index, off_t size, int random,
					t_st *begin, t_st *end);
	void			add_to_directory(
					struct s_floppy *pfs,
					char *name, char *ext, int dir_index,
					int random, struct stat *pstat,
					t_st *begin, t_st *end, int wp);
	void			modify_random_file(char *path, struct stat *pstat,
					t_st begin);
	int			is_flex_filename(char *filename,
					char *name, char *ext);
	void			check_for_delete(int dir_index, s_dir_sector *buffer);
	void			check_for_extend(int dir_index, s_dir_sector *buffer);
	void			check_for_rename(int dir_index, s_dir_sector *pb);
	int			check_for_new_file(int dir_index, s_dir_sector *pd);
	int			extend_directory(struct s_floppy *pfs,
					int index, s_dir_sector *pdb);
	int			set_file_time(char *ppath,
					Byte month, Byte day, Byte year);
	t_st			*link_address(void);
	int			last_of_free_chain(Byte tr, Byte sr);
	void			read_nafs_sector(void);
	void			write_nafs_sector(void);
	int			index_of_new_file(Byte track, Byte sector);
	void			change_file_id(int index, int old_id, int new_id);
public:
	int			update(void);

#endif // NAFS
public:

	virtual void		 reset(void);
	virtual Byte		 readstatus(Word offset);
	virtual void		 writedrisel(Word offset, Byte val);
				 E2floppy(Mc6809* cpu, char *drive[]);
				~E2floppy(void);
};

#endif /* #ifndef __fromflex__ */
#endif /* __e2floppy_h__ */

