/*
    gui.cc


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include "gui.h"
#include "e2floppy.h"
#include "e2video.h"

#ifdef XPM
#include "author.xpm"
#endif

#ifndef XLIB
int menu_id[] = {11, 12, 21, 22, 23, 24, 31, 32};
int button_id[] = {1, 2, 11, 21, 22, 23, 24, 25, 26, 27, 28, 31, 32, 41, 51,
		52, 53, 54, 55, 56, 57, 58};

/* ARGSUSED */
void expose(Widget w, XEvent *pevent, String *pparams, Cardinal *pnparams)
{
	ggui->c_expose(pevent);
}

#ifdef XPM
/* ARGSUSED */
void author_expose(Widget w, XEvent *pevent, String *pparams, Cardinal *pnparams)
{
	ggui->a_expose(w, pevent);
}
#endif

/* ARGSUSED */
void keyPress(Widget w, XEvent *pevent, String *pparams, Cardinal *pnparams)
{
	ggui->c_keyPress(pevent);
}

/* ARGSUSED */
void popup_no_resources(Widget w, XEvent *pevent, String *pparams,
	Cardinal *pnparams)
{
	ggui->popup_message("");
	ggui->set_exit();
}

/* ARGSUSED */
void menuCallback(Widget w, XtPointer client_data, XtPointer call_data)
{
	switch (*(int *)client_data) {
#ifdef NAFS
		case 11: ggui->update_nafs(); break;
#endif
		case 12: ggui->set_new_state(EXIT); break;
		case 21: ggui->set_new_state(RUN);  break;
		case 22: ggui->set_new_state(STOP); break;
		case 23: ggui->set_new_state(RESET); break;
		case 24: ggui->popup_cpu(); break;
		case 31: ggui->popup_help(); break;
		case 32: ggui->popup_about(); break;
	} // switch
} // menuCallback

/* ARGSUSED */
void buttonCallback(Widget w, XtPointer client_data, XtPointer call_data)
{
	switch (*(int *)client_data) {
		case 1: ggui->popdown_about(); break;  // Ok in about
#ifdef XPM
		case 2: ggui->popdown_about();
			ggui->popup_author(w);  break;  // Author in about
		case 11: ggui->popdown_author(w); break; // Ok in author
#endif
		case 21: ggui->set_new_state(RUN); break; // Run in cpu
		case 22: ggui->set_new_state(STOP); break; // Stop in cpu
		case 23: ggui->set_new_state(STEP); break; // Step in cpu
		case 24: ggui->set_new_state(NEXT); break; // Next in cpu
		case 25: ggui->set_new_state(RESET); break; // Reset in cpu
		case 26: ggui->popup_bp(0); break; // Bp1 in cpu
		case 27: ggui->popup_bp(1); break; // Bp2 in cpu
		case 28: ggui->popdown_cpu(); break; // Ok in cpu
		case 31: ggui->popdown_bp(); break; // Ok in bp
		case 32: ggui->clear_bp(); break; // Clear in bp
		case 41: ggui->popdown_message(); break; // Ok in messageBox
		case 51: ggui->popdown_help(); break; // Ok in Help 
		case 52: case 53: case 54: case 55: case 56: case 57: case 58:
			ggui->display_man_page(w); break; // "manpage" in Help 
	} // switch
} // menuCallback
#endif

Gui::Gui(
	Mc6809* x_cpu,
	Memory* x_memory,
	Inout*  x_io,
	E2video* x_video,
	int	argc,
	char 	*argv[],
	char    *color,
	int      inverse)
{
	cpu		= x_cpu;
	memory		= x_memory;
	io		= x_io;
	e2video		= x_video;
	program_name	= argv[0];
	exit_flag	= 0;

	e2video->reset();
#ifdef XLIB
	display = XOpenDisplay((char *)NULL);
	if (display == NULL) {
		fprintf(stderr, "can't open %s\n", XDisplayName((char *)NULL));
		exit(1);
	};
	memory->initialize_byte_conv_table(
		XImageByteOrder(display) == LSBFirst);
	screen = DefaultScreenOfDisplay(display);
	initialize_e2window(argc, argv, inverse, color);
	initialize_cpu_window();
#else
	initialize_e2window(argc, argv, inverse, color);
#endif
} 

Gui::~Gui(void)
{
#ifdef XLIB
	XDestroyWindow(display, cpu_window);
	XDestroyWindow(display, e2window);
#else
	XtCloseDisplay(XtDisplay(e2toplevel));
#endif
}

void Gui::set_exit(void)
{
	exit_flag = 1;
}

void Gui::update_block(int block_number)
{
if (e2video->are_blocks_to_update() || memory->changed[block_number]) {
	if (!(e2video->vico1 & 0x02)) {
		if (block_number == e2video->divided_block) {
			// first half display on the bottom of the window
#ifdef XLIB
			XPutImage(display, e2window,
#else
			XPutImage(XtDisplay(e2screen), XtWindow(e2screen),
#endif
				e2gc, image[e2video->vico1][block_number],
				0, 0, 0, WINDOWHEIGHT -
				   (e2video->vico2 % BLOCKHEIGHT),
				BLOCKWIDTH, e2video->vico2 % BLOCKHEIGHT);
			// second half display on the top of window
#ifdef XLIB
			XPutImage(display, e2window,
#else
			XPutImage(XtDisplay(e2screen), XtWindow(e2screen),
#endif
				e2gc, image[e2video->vico1][block_number],
				0, e2video->vico2 % YBLOCKS, 0, 0,
				BLOCKWIDTH, BLOCKHEIGHT -
				   (e2video->vico2 % BLOCKHEIGHT));
		} else {
#ifdef XLIB
			XPutImage(display, e2window,
#else
			XPutImage(XtDisplay(e2screen), XtWindow(e2screen),
#endif
				e2gc, image[e2video->vico1][block_number],
				0, 0, 0,
				(block_number * BLOCKHEIGHT + WINDOWHEIGHT -
				   e2video->vico2) % WINDOWHEIGHT,
				BLOCKWIDTH, BLOCKHEIGHT);
		} // else
		memory->changed[block_number] = 0;
	} else {
#ifdef XLIB
			XPutImage(display, e2window,
#else
			XPutImage(XtDisplay(e2screen), XtWindow(e2screen),
#endif
				e2gc, image2, 0, 0, 0,
				block_number * BLOCKHEIGHT,
				BLOCKWIDTH, BLOCKHEIGHT);
	} // else
} // if
} // update_block

void Gui::update(void)
{
	static int event_loop_counter = 0;
	static int display_block = 0;

	update_block(display_block);
	display_block = ++display_block % YBLOCKS;
	if (event_loop_counter++ == 25) {
		event_loop_counter = 0;
		event_queue();
	} // if
}

void Gui::set_new_state(unsigned int new_state)
{
	if (new_state != NO_CHANGE)
		cpu->set_new_state(new_state);
}

#ifndef XLIB
#ifdef NAFS
void Gui::update_nafs(void)
{
	if (!memory->fdc->update())
		popup_message("can't update file system\nThere are open files\n");
}
#endif // ifdef NAFS
#endif // ifndef XLIB

#ifndef XLIB
#ifdef XPM
void Gui::popup_author(Widget w)
{
	XtPopup(authorframe, XtGrabExclusive);
}

void Gui::a_expose(Widget w, XEvent *pevent)
{
	static int errorStatus = XpmSuccess;
	XpmAttributes xpmattr;

	if (authorpixmap == 0 && errorStatus == XpmSuccess) {
		xpmattr.colormap  = DefaultColormapOfScreen(XtScreen(w));
		xpmattr.valuemask = XpmColormap; 
		errorStatus = XpmCreatePixmapFromData(XtDisplay(w),
			XtWindow(w), author, &authorpixmap, NULL, &xpmattr);
		XpmFreeAttributes(&xpmattr);
		if (errorStatus != XpmSuccess)
			popup_message("\
sorry, unable to create image\n\
of author. (image need 62\n\
color entries)");
	}
	if (errorStatus == XpmSuccess)
		XCopyArea(XtDisplay(w), authorpixmap, XtWindow(w), e2gc,
			pevent->xexpose.x,     pevent->xexpose.y,
			pevent->xexpose.width, pevent->xexpose.height,
			pevent->xexpose.x,     pevent->xexpose.y);
}

void Gui::popdown_author(Widget w)
{
	XtPopdown(authorframe);
	if (authorpixmap) {
		XFreePixmap(XtDisplay(authorframe), authorpixmap);
		authorpixmap = 0;
	}
}
#endif

void Gui::popup_about(void)
{
	XtPopup(aboutframe, XtGrabExclusive);
}

void Gui::popdown_about(void)
{
	XtPopdown(aboutframe);
}

void Gui::popup_help(void)
{
	XtPopup(helpframe, XtGrabNone);
}

void Gui::popdown_help(void)
{
	XtPopdown(helpframe);
}

void Gui::popup_message(char *pmessage)
{
	if (*pmessage)
		XtVaSetValues(messagetext, XtNstring, pmessage, NULL);
	XtPopup(messageframe, XtGrabExclusive);
}

void Gui::popdown_message(void)
{
	XtPopdown(messageframe);
	if (exit_flag)
		set_new_state(EXIT);
}
#endif

int Gui::translate_to_ascii(XKeyEvent *pevent)
{
	KeySym keysym;
	char charString[2];
	XComposeStatus compose;
	int count;

	// first check for control character
	keysym = XLookupKeysym(pevent, 0);
//	fprintf(stderr, "%X\n", keysym); only for tests
	if (pevent->state & ControlMask) {
		switch(keysym) {
			case XK_a: return 0x01;
			case XK_b: return 0x02;
			case XK_c: return 0x03;
			case XK_d: return 0x04;
			case XK_e: return 0x05;
			case XK_f: return 0x06;
			case XK_g: return 0x07;
			case XK_h: return 0x08;
			case XK_i: return 0x09;
			case XK_j: return 0x0a;
			case XK_k: return 0x0b;
			case XK_l: return 0x0c;
			case XK_m: return 0x0d;
			case XK_n: return 0x0e;
			case XK_o: return 0x0f;
			case XK_p: return 0x10;
			case XK_q: return 0x11;
			case XK_r: return 0x12;
			case XK_s: return 0x13;
			case XK_t: return 0x14;
			case XK_u: return 0x15;
			case XK_v: return 0x16;
			case XK_w: return 0x17;
			case XK_x: return 0x18;
			case XK_y: return 0x19;
			case XK_z: return 0x1a;
			case XK_8:         return 0x1b;
			case XK_ssharp:    return 0x1c;
			case XK_9:         return 0x1d;
			case XK_asciicircum:return 0x1e;
			case XK_minus:     return 0x1f;
			case XK_Return:    return 0x0d;
			case XK_Escape:    return 0x1b;
			case XK_Tab:       return 0x09;
			case XK_Delete:
			case XK_BackSpace: return 0x08;
			case XK_Break:
			case XK_Pause:     if (cpu) cpu->set_nmi(); return -1;

			case XK_Home:
			case XK_KP_7:	   return 0xb7;
			case XK_Up:
			case XK_KP_8:	   return 0xb8;
			case XK_Prior:
			case XK_KP_9:	   return 0xb9;
			case XK_Left:
			case XK_KP_4:	   return 0xb4;
			case XK_Begin:
			case XK_KP_5:	   return 0xb5;
			case XK_Right: 
			case XK_KP_6:	   return 0xb6;
			case XK_KP_1:	   return 0xb1;
			case XK_Down:
			case XK_KP_2:	   return 0xb2;
			case XK_Next:
			case XK_KP_3:	   return 0xb3;
			default:           return -1;
		} // switch
	} // if
	if (pevent->state & ShiftMask) {
		switch(keysym) {
			case XK_F1:    return 0xca;
			case XK_F2:    return 0xcb;
			case XK_F3:    return 0xcc;
			case XK_F4:    return 0xcd;
			case XK_F5:    return 0xce;
			case XK_F6:    return 0xcf;
			case XK_F7:    return 0xd0;
			case XK_F8:    return 0xd1;
			case XK_F9:    return 0xd2;

			case XK_Break:
			case XK_Pause:     if (cpu) cpu->set_nmi(); return -1;
			case XK_Delete:
			case XK_BackSpace: return 0x7f;

			case XK_Home:
			case XK_KP_7:	   return 0xe7;
			case XK_Up:
			case XK_KP_8:	   return 0xe8;
			case XK_Prior:
			case XK_KP_9:	   return 0xe9;
			case XK_Left:
			case XK_KP_4:	   return 0xe4;
			case XK_Begin:
			case XK_KP_5:	   return 0xe5;
			case XK_Right: 
			case XK_KP_6:	   return 0xe6;
			case XK_KP_1:	   return 0xe1;
			case XK_Down:
			case XK_KP_2:	   return 0xe2;
			case XK_Next:
			case XK_KP_3:	   return 0xe3;
		} // switch
	} // if
	switch (keysym) {
		// main keyboard
		case XK_Return:    return 0x0d;
		case XK_Escape:    return 0x1b;
		case XK_Tab:       return 0x09;
		case XK_space:     return ' ';
		case XK_Delete:
		case XK_BackSpace: return 0x08;
		case XK_Break:
		case XK_Pause:     return 0x03;

		case XK_F1:        return 0xc0;
		case XK_F2:        return 0xc1;
		case XK_F3:        return 0xc2;
		case XK_F4:        return 0xc3;
		case XK_F5:        return 0xc4;
		case XK_F6:        return 0xc5;
		case XK_F7:        return 0xc6;
		case XK_F8:        return 0xc7;
		case XK_F9:        return 0xc8;
		case XK_F10:       return 0xc9;
		case XK_F11:       return 0xca;
		case XK_F12:       return 0xcb;
		
		case XK_Home:
		case XK_KP_7:	   return 0xf7;
		case XK_Up:
		case XK_KP_8:	   return 0xf8;
		case XK_Prior:
		case XK_KP_9:	   return 0xf9;
		case XK_Left:
		case XK_KP_4:	   return 0xf4;
		case XK_Begin:
		case XK_KP_5:	   return 0xf5;
		case XK_Right: 
		case XK_KP_6:	   return 0xf6;
		case XK_KP_1:	   return 0xf1;
		case XK_Down:
		case XK_KP_2:	   return 0xf2;
		case XK_Next:
		case XK_KP_3:	   return 0xf3;
		case XK_KP_Decimal:return ',';
		case XK_KP_Add:    return '+';
		case XK_KP_Subtract:return '-';
		case XK_KP_Enter:  return 0x0d;
	} // switch
	count = XLookupString(pevent, charString, 2,
				&keysym, &compose);
	if (count == 1) {
		switch(charString[0]) {
			case '': return '^';
			case '': return '~';
			case '': return '[';
			case '': return '\\';
			case '': return ']';
			case '': return '{';
			case '': return '|';
			case '': return '}';
			default: return charString[0];
		} // switch
	}
	return -1;
} // translate_to_ascii

void Gui::event_queue(void)
{
	XEvent event;
	Window root, child;
	int newX, newY;
	int winX, winY;
	unsigned int newButtonMask;
	int percent;
#ifdef XLIB
	KeySym keysym;
	long event_mask;
	int i, key = 0;

	event_mask = ButtonPressMask | ExposureMask | KeyPressMask;
	if (XCheckWindowEvent(display, cpu_window, event_mask, &event))
		switch (event.type) {
		   case NoExpose:
		        break;
		   case ButtonPress:
		   	   process_button_input(&event);
			   break;
		   case KeyPress: 
			keysym = XLookupKeysym(&event.xkey,
				(event.xkey.state & ShiftMask) ?
				 ShiftMapIndex : 0);
			for (i=0; i < 2; i++) {
				if (bp_input[i])
					process_bp_input(keysym, i);
			}
			update_1_second(NO_CHANGE);
			break;
		   case Expose:
			update_all();
			break;
		} // switch
	event_mask = ExposureMask | KeyPressMask;
	if (XCheckWindowEvent(display, e2window, event_mask, &event))
		switch (event.type) {
		   case KeyPress:
		       if (!io->key_buffer_full() &&
			   (key = translate_to_ascii(&event.xkey)) >= 0)
				io->put_ch(key);
			break;
		   case Expose:
			e2video->init_blocks_to_update();
			break;
		} // switch
#else
	// process all queued events
	while (XtAppPending(context)) {
//		XtAppNextEvent(context, &event);
//		XtDispatchEvent(&event);
		XtAppProcessEvent(context, XtIMXEvent);
	} // while
#endif

#ifdef XLIB
	if (XQueryPointer(display, e2window,
#else
	if (XQueryPointer(XtDisplay(e2screen), XtWindow(e2screen),
#endif
		&root, &child, &newX, &newY,
		&winX, &winY, &newButtonMask)) {
		io->deltaX = newX - oldX;
		io->deltaY = newY - oldY;
		io->buttonMask = newButtonMask;
		oldX = newX;
		oldY = newY;
	} // if
	if (io->is_bell_request(&percent))
#ifdef XLIB
		XBell(display, percent);
#else
		XBell(XtDisplay(e2screen), percent);
#endif
	if (io->poll())
		memory->pia1->activeTransition(CA1);
}

#ifndef XLIB
/* ARGSUSED */
void Gui::c_expose(XEvent *pevent)
{

	if (pevent->xexpose.count == 0 && XtIsRealized(e2screen))
		e2video->init_blocks_to_update();
} // c_expose

void Gui::c_keyPress(XEvent *pevent)
{
	int key;

	if ((key = translate_to_ascii(&pevent->xkey)) >= 0)
		io->put_ch(key);
} // keyPress
#endif

#ifdef XLIB
void Gui::initialize_e2window(int argc, char *argv[], int inverse, char *color)
{
	unsigned int		i, j;
	Visual		       *visual;
	XGCValues		gcv;
	XEvent			event;
	XSetWindowAttributes	attrs;
	XColor			xcolor, exact_color;
	XSizeHints		xsh;
	XClassHint		xch;

	oldX       = 0;
	oldY       = 0;
	exit_flag  = 0;

	// handle some default values
	if (!*color)
		color = "green";

	attrs.event_mask		= KeyPressMask | ExposureMask;

	e2window = XCreateWindow(display, RootWindowOfScreen(screen), 0, 0,
		WINDOWWIDTH, WINDOWHEIGHT, 0, CopyFromParent,
		InputOutput, (Visual *)CopyFromParent,
		CWEventMask, &attrs);

	if (!XAllocNamedColor(display, DefaultColormapOfScreen(screen),
	   color, &xcolor, &exact_color))
		xcolor.pixel = WhitePixelOfScreen(screen);
	if (!inverse) {
		gcv.foreground = xcolor.pixel;
		gcv.background = BlackPixelOfScreen(screen);
	} else {
		gcv.foreground = BlackPixelOfScreen(screen);
		gcv.background = xcolor.pixel;
	};
	e2gc = XCreateGC(display, e2window, GCForeground | GCBackground, &gcv);

	xsh.min_width  = xsh.max_width  = WINDOWWIDTH;
	xsh.min_height = xsh.max_height = WINDOWHEIGHT;
	xsh.flags = PMinSize | PMaxSize;
	XSetStandardProperties(display, e2window, "flexemu", "flexemu",
		None, NULL, 0, &xsh);
		
	xch.res_name = "e2video";
	xch.res_class = "E2video";
	XSetClassHint(display, e2window, &xch);

	visual = DefaultVisualOfScreen(screen);
	for (j = 0; j < YBLOCKS; j++)
		for (i = 0; i < 2; i++) {
			image[i][j] = XCreateImage(display, visual,
			1, XYBitmap, 0,
			(char *)(memory->video_ram + j * YBLOCK_SIZE +
				i * VIDEORAM_SIZE * 6),
				BLOCKWIDTH, BLOCKHEIGHT, 32, 0);
		} // for
	for (i = 0; i < YBLOCK_SIZE; i++)
		unused_block[i] = 0xff; // always display as all one's
	image2 = XCreateImage(display, visual, 1, XYBitmap, 0,
		(char *)unused_block, BLOCKWIDTH, BLOCKHEIGHT, 32, 0);

	// show window on display
	XMapWindow(display, e2window);
	// wait, until window is visible
	XWindowEvent(display, e2window, ExposureMask, &event);
	e2video->init_blocks_to_update();
}
#else
void Gui::initialize_e2window(int argc, char *argv[], int inverse, char *color)
{
	Widget w;

	w = create_main_view(argc, argv);
	create_message_dialog(w);
	create_about_dialog(w);
	create_help_view(w);
#ifdef XPM
	create_author_dialog(w);
#endif
	create_cpu_view(w);
	create_bp_dialog(w);
	initialize_after_create(w, inverse, color);
	manage_widget(w);
	initialize_after_open(w);
	e2toplevel = w;
} // initialize_e2window
#endif

#ifndef XLIB
Widget Gui::create_main_view(int argc, char *argv[])
{
	Widget mainview;

	// create widget tree:
#ifdef sun
	mainview = XtVaAppInitialize(&context, "Flexemu",
		opts, 0, (unsigned int *)&argc, argv,
		(const char **)fallback_resources, NULL);
#else
	mainview = XtVaAppInitialize(&context, "Flexemu",
		opts, 0, &argc, argv, fallback_resources, NULL);
#endif

	XtAppAddConverter(context, XtRString, XtROrientation,
		XmuCvtStringToOrientation, NULL, 0);
	form = XtVaCreateManagedWidget("form", formWidgetClass, mainview,
		NULL);
	menubar = XtVaCreateManagedWidget("menuBar", formWidgetClass, form,
		NULL);
	button1 = XtVaCreateManagedWidget("menuButton1", menuButtonWidgetClass,
		menubar, NULL);
	button2 = XtVaCreateManagedWidget("menuButton2", menuButtonWidgetClass,
		menubar, NULL);
	button3 = XtVaCreateManagedWidget("menuButton3", menuButtonWidgetClass,
		menubar, XtNright, XtChainRight, NULL);
	e2screen = XtVaCreateManagedWidget("screen", coreWidgetClass, form,
		XtNwidth, (XtArgVal)WINDOWWIDTH,
		XtNheight, (XtArgVal)WINDOWHEIGHT, NULL);
	menu1 = XtVaCreatePopupShell("menu1", simpleMenuWidgetClass, button1,
		NULL);
	menu2 = XtVaCreatePopupShell("menu2", simpleMenuWidgetClass, button2,
		NULL);
	menu3 = XtVaCreatePopupShell("menu3", simpleMenuWidgetClass, button3,
		NULL);
#ifdef NAFS
	entry11 = XtVaCreateManagedWidget("menuEntry11", smeBSBObjectClass,
		menu1, NULL);
	line1 = XtVaCreateManagedWidget("line1", smeLineObjectClass, menu1, NULL);
#endif
	entry12 = XtVaCreateManagedWidget("menuEntry12", smeBSBObjectClass,
		menu1, NULL);
	entry21 = XtVaCreateManagedWidget("menuEntry21", smeBSBObjectClass,
		menu2, NULL);
	entry22 = XtVaCreateManagedWidget("menuEntry22", smeBSBObjectClass,
		menu2, NULL);
	entry23 = XtVaCreateManagedWidget("menuEntry23", smeBSBObjectClass,
		menu2, NULL);
	line2 = XtVaCreateManagedWidget("line2", smeLineObjectClass, menu2, NULL);
	entry24 = XtVaCreateManagedWidget("menuEntry24", smeBSBObjectClass,
		menu2, NULL);
	entry31 = XtVaCreateManagedWidget("menuEntry31", smeBSBObjectClass,
		menu3, NULL);
	entry32 = XtVaCreateManagedWidget("menuEntry32", smeBSBObjectClass,
		menu3, NULL);

	// add some menu callbacks:
#ifdef NAFS
	XtAddCallback(entry11, XtNcallback, menuCallback, &menu_id[0]);
#endif
	XtAddCallback(entry12, XtNcallback, menuCallback, &menu_id[1]);
	XtAddCallback(entry21, XtNcallback, menuCallback, &menu_id[2]);
	XtAddCallback(entry22, XtNcallback, menuCallback, &menu_id[3]);
	XtAddCallback(entry23, XtNcallback, menuCallback, &menu_id[4]);
	XtAddCallback(entry24, XtNcallback, menuCallback, &menu_id[5]);
	XtAddCallback(entry31, XtNcallback, menuCallback, &menu_id[6]);
	XtAddCallback(entry32, XtNcallback, menuCallback, &menu_id[7]);
	return mainview;
}  // create_main_view


void Gui::create_about_dialog(Widget parent)
{
	char			*aboutstring;

	aboutstring = XtMalloc(sizeof(VERSION) + sizeof(HEADER1) +
		sizeof(HEADER2));

	if (aboutstring != NULL) {
		strcpy(aboutstring, HEADER1);
       		strcat(aboutstring, VERSION);
       		strcat(aboutstring, HEADER2);
       	} else {
       		aboutstring = "not enough memory";
       	}

	aboutframe = XtVaCreatePopupShell("about",
		transientShellWidgetClass, parent, NULL);
	aboutform = XtVaCreateManagedWidget("aboutForm",
		formWidgetClass, aboutframe, NULL);
	abouttext = XtVaCreateManagedWidget("aboutText", asciiTextWidgetClass,
		aboutform,
		XtNstring, (XtArgVal)aboutstring,
		XtNlength, (XtArgVal)(strlen(aboutstring)+1),
		NULL);
	aboutbutton = XtVaCreateManagedWidget("aboutButton",
		commandWidgetClass, aboutform,
#ifdef XPM
		XtNhorizDistance, (XtArgVal)80, NULL);
#else
		XtNhorizDistance, (XtArgVal)100, NULL);
#endif

#ifdef XPM
	morebutton = XtVaCreateManagedWidget("moreButton",
		commandWidgetClass, aboutform, NULL);
#endif

	// define some button callbacks:
	XtAddCallback(aboutbutton, XtNcallback, buttonCallback, &button_id[0]);
#ifdef XPM
	XtAddCallback(morebutton, XtNcallback, buttonCallback, &button_id[1]);
#endif
	XtFree(aboutstring);

} // create_about_dialog

// the returned string must be released with XtFree()!!

String Gui::get_man_page(char *pman_page_name)
{
	String	 phelpstring, ptmp;
	char	*p, ch1, ch2;
	char	*perr = "can't find man page ";
	FILE	*fp;
	char	 buffer[32];
	int	 count, error;

	error = 0;
	count = 0;
	ch1 = '\0';
	phelpstring = XtMalloc(8192);
	sprintf((char *)buffer, "man %s 2>&1", pman_page_name);
	if ((fp = popen(buffer , "r")) != NULL) {
		p = phelpstring;
		if (!feof(fp))
			ch1 = fgetc(fp);
		while (!feof(fp)) {
			ch2 = fgetc(fp);
			if (ch2 != 8 && ch1 != 8) { // 8 means ^H
				*(p++) = ch1;
				count++;
			}
			ch1 = ch2;
			if ((count & 0x1fff) == 0x1ffe) {
				if ((ptmp = XtMalloc(count + 8192)) != NULL) {
					memcpy(ptmp, phelpstring, count);
					p = ptmp + count;
					XtFree(phelpstring);
					phelpstring = ptmp;
				} else
					error = 1;
			} // if
		} // while
		// there are at least two bytes free in the stringbuffer
		if (ch1 >= ' ') 
			*(p++) = ch1;
		*p = '\0';
		if (pclose(fp) < 0)
			error = 2;
	} else
		error = 3;
	if (error || strlen(phelpstring) == 0) {
		strcpy(phelpstring, perr);
		strcat(phelpstring, pman_page_name);
	}  // if
	return phelpstring;
} // get_man_page

void Gui::create_help_view(Widget parent)
{
	String phelpstring;

	phelpstring = get_man_page("flexemu");
	helpframe = XtVaCreatePopupShell("help",
		transientShellWidgetClass, parent, NULL);
	helpform = XtVaCreateManagedWidget("helpForm",
		formWidgetClass, helpframe, NULL);
	helptext = XtVaCreateManagedWidget("helpText", asciiTextWidgetClass,
		helpform,
		XtNstring,	     (XtArgVal)phelpstring,
		XtNscrollVertical,   (XtArgVal)XawtextScrollWhenNeeded,
		XtNscrollHorizontal, (XtArgVal)XawtextScrollWhenNeeded,
		NULL);
	helpbutton = XtVaCreateManagedWidget("helpButton",
		commandWidgetClass, helpform, NULL);
	helpbutton1 = XtVaCreateManagedWidget("helpButton1",
		commandWidgetClass, helpform, NULL);
	helpbutton2 = XtVaCreateManagedWidget("helpButton2",
		commandWidgetClass, helpform, NULL);
	helpbutton3 = XtVaCreateManagedWidget("helpButton3",
		commandWidgetClass, helpform, NULL);
	helpbutton4 = XtVaCreateManagedWidget("helpButton4",
		commandWidgetClass, helpform, NULL);
	helpbutton5 = XtVaCreateManagedWidget("helpButton5",
		commandWidgetClass, helpform, NULL);
	helpbutton6 = XtVaCreateManagedWidget("helpButton6",
		commandWidgetClass, helpform, NULL);
	helpbutton7 = XtVaCreateManagedWidget("helpButton7",
		commandWidgetClass, helpform, NULL);
	XtFree(phelpstring);

	// define some button callbacks:
	XtAddCallback(helpbutton,  XtNcallback, buttonCallback, &button_id[14]);
	XtAddCallback(helpbutton1, XtNcallback, buttonCallback, &button_id[15]);
	XtAddCallback(helpbutton2, XtNcallback, buttonCallback, &button_id[16]);
	XtAddCallback(helpbutton3, XtNcallback, buttonCallback, &button_id[17]);
	XtAddCallback(helpbutton4, XtNcallback, buttonCallback, &button_id[18]);
	XtAddCallback(helpbutton5, XtNcallback, buttonCallback, &button_id[19]);
	XtAddCallback(helpbutton6, XtNcallback, buttonCallback, &button_id[20]);
	XtAddCallback(helpbutton7, XtNcallback, buttonCallback, &button_id[21]);
} // create_help_view

void Gui::display_man_page(Widget w)
{
	String plabel, phelpstring;
	char label[16];

	// the label of the button is the name of the man-page
	// so it can simply be changed by editing the resource file

	XtVaGetValues(w, XtNlabel, &plabel, NULL);
	strcpy((char *)&label, plabel);
	label[0] = tolower(label[0]); // lower first character
	phelpstring = get_man_page(label);
	XtVaSetValues(helptext, XtNstring, (XtArgVal)phelpstring, NULL);
	XtFree(phelpstring);
} // display_man_page


#ifdef XPM
void Gui::create_author_dialog(Widget parent)
{
	authorpixmap = 0;
	authorframe = XtVaCreatePopupShell("author",
		transientShellWidgetClass, parent, NULL);
	authorform = XtVaCreateManagedWidget("authorForm",
		formWidgetClass, authorframe, NULL);
	authorwidget = XtVaCreateManagedWidget("authorWidget",
		coreWidgetClass, authorform, NULL);
	authorbutton = XtVaCreateManagedWidget("authorButton",
		commandWidgetClass, authorform, NULL);

	// define some button callbacks:
	XtAddCallback(authorbutton, XtNcallback, buttonCallback, &button_id[2]);
} // create_author_dialog
#endif

void Gui::create_message_dialog(Widget parent)
{
	messageframe = XtVaCreatePopupShell("message",
		transientShellWidgetClass, parent, NULL);
	messageform = XtVaCreateManagedWidget("messageForm",
		formWidgetClass, messageframe, NULL);
	messagetext = XtVaCreateManagedWidget("messageText",
		asciiTextWidgetClass, messageform, NULL);
	messagebutton = XtVaCreateManagedWidget("messageButton",
		commandWidgetClass, messageform, NULL);

	// define some button callbacks:
	XtAddCallback(messagebutton, XtNcallback, buttonCallback, &button_id[13]);
} // create_message_dialog

void Gui::initialize_after_create(Widget w, int inverse, char *color)
{
	unsigned int		i, j;
	Display		       *display;
	Screen		       *screen;
	Visual		       *visual;
	XGCValues		gcv;
	XColor			xcolor, exact_color;
	static XtActionsRec	actions[] = {
#ifdef XPM
					{"author_expose", (XtActionProc)author_expose},
#endif
					{"expose",   (XtActionProc)expose},
					{"keyPress", (XtActionProc)keyPress},
					{"convert", (XtActionProc)keyPress},
					{"popup_no_resources", (XtActionProc)popup_no_resources}
				};


	oldX       = 0;
	oldY       = 0;

	display = XtDisplay(w);
	screen  = XtScreen(w);
	visual  = DefaultVisualOfScreen(screen);

	XtAppAddActions(context, actions, XtNumber(actions));

		memory->initialize_byte_conv_table(
			XImageByteOrder(display) == LSBFirst);

	if (!*color)
		color = "green";
	if (!XAllocNamedColor(display, DefaultColormapOfScreen(screen),
		color, &xcolor, &exact_color))
		xcolor.pixel = WhitePixelOfScreen(screen);
	if (!inverse) {
		gcv.foreground = xcolor.pixel;
		gcv.background = BlackPixelOfScreen(screen);
	} else {
		gcv.foreground = BlackPixelOfScreen(screen);
		gcv.background = xcolor.pixel;
	}
	e2gc = XtGetGC(e2screen, GCForeground | GCBackground, &gcv);
	XtVaSetValues(e2screen, XtNbackground, gcv.background, NULL);

	// initialize images
	for (j = 0; j < YBLOCKS; j++)
		for (i = 0; i < 2; i++) {
			image[i][j] = XCreateImage(display,
			visual, 1, XYBitmap, 0,
			(char *)(memory->video_ram + j * YBLOCK_SIZE +
				i * VIDEORAM_SIZE * 6),
				BLOCKWIDTH, BLOCKHEIGHT, 32, 0);
		} // for
	for (i = 0; i < YBLOCK_SIZE; i++)
		unused_block[i] = 0xff; // always display as all one's
	image2 = XCreateImage(display, visual, 1, XYBitmap, 0,
		(char *)unused_block, BLOCKWIDTH, BLOCKHEIGHT, 32, 0);

	// needed for propper expose and key press:
	XtInstallAllAccelerators(e2screen, form);
#ifdef XPM
	XtInstallAllAccelerators(authorwidget, authorframe);
#endif
}

void Gui::manage_widget(Widget w)
{
	XtRealizeWidget(w);
	XtMapWidget(w);
	e2video->init_blocks_to_update();	// update main view
	update_all();				// update cpu view
}  // manage widget


void Gui::initialize_after_open(Widget w)
{
	Dimension	width, height;
	XSizeHints	xsh;

	// don't allow to resize window:
	XtVaGetValues(w, XtNwidth, &width, XtNheight, &height, NULL);
	xsh.max_width  = xsh.min_width  = width;
	xsh.max_height = xsh.min_height = height;
	xsh.flags      = PMaxSize | PMinSize;
	XSetStandardProperties(XtDisplay(w), XtWindow(w),
		"flexemu", "flexemu", None, NULL, 0, &xsh);

}
#endif

