/*
    gui.h


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/



#ifndef __gui_h__
#define __gui_h__

#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>

#ifndef XLIB
#ifdef XPM
#include <X11/xpm.h>
#endif
#include <X11/Xaw/Simple.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Toggle.h>
#include <X11/RectObj.h>
#include <X11/Xmu/Converters.h>
#include <X11/Shell.h>
#endif

#include "misc.h"
#include "e2.h"
#include "e2video.h"
#include "mc6809.h"
#include "pia1.h"
#include "inout.h"

#define VERSION "V 0.80"
#define HEADER1 "                 FLEXEMU\n\
     an MC6809 emulator running FLEX\n                 "
#define HEADER2 "\n\n\
     Copyright (C) 1997 W. Schwotzer\n\n\
FLEXEMU comes with ABSOLUTELY NO WARRANTY.\n\
This is free software, and you are welcome\n\
    to redistribute it under certain\n\
    conditions. Look at file COPYING.\n\n\
          woschwot@si.bosch.de"

#ifndef XLIB
extern int button_id[];
extern void buttonCallback(Widget w, XtPointer client_data, XtPointer call_data);
extern String fallback_resources[];
#endif

#ifdef XLIB
#define BUTTON_X	30
#define BUTTON_RUN_Y	1
#define BUTTON_STOP_Y	2
#define BUTTON_STEP_Y	3
#define BUTTON_NEXT_Y	4
#define BUTTON_RESET_Y	5
#define BUTTON_EXIT_Y	6
#endif

#define LINE_SIZE	39	// linesize in cpu view

class Gui {

#ifdef XLIB
// breakpoint support
protected:
	Byte		bp_input[2];
#else
	unsigned int	which_bp;
#endif

protected:

// X11 stuff
#ifdef XLIB
	Screen		*screen;
	Display		*display;
	Window		 cpu_window;
	Pixmap		 pix;
	GC		 cpu_gc, inv_gc;
	XFontStruct	*fs;
	int		 ww, wh;
#endif
	Window		 e2window;
	GC		 e2gc;
	char 		*program_name;
	unsigned char 	 unused_block[YBLOCK_SIZE];
	XImage 		*image[2][YBLOCKS];
	XImage 		*image2;

// Xtoolkit stuff
#ifndef XLIB
	Widget 		 e2toplevel, form, menubar, e2screen;
	Widget		 button1, button2, button3, menu1, menu2, menu3;
	Widget		 entry11, entry12,
			 entry21, entry22, entry23, entry24;
	Widget		 entry31, entry32, line1, line2;
	Widget		 messageframe, messageform, messagetext, messagebutton;
	Widget		 helpframe, helpform, helptext, helpbutton;
	Widget		 helpbutton1, helpbutton2, helpbutton3, helpbutton4;
	Widget		 helpbutton5, helpbutton6, helpbutton7;
	Widget		 aboutframe, aboutform, abouttext, aboutbutton;
	Widget		 cpuframe, cpuform, cputext;
	Widget		 nextbutton, stepbutton, runbutton, stopbutton;
	Widget		 bp1button, bp2button, cpubutton, resetbutton;
	Widget		 bpframe, bpform, bptext, bpbutton, clearbutton;
#ifdef XPM
	Widget		 morebutton;
	Widget		 authorframe, authorform, authorwidget, authorbutton;
	Pixmap		 authorpixmap;
#endif
	XtAppContext	 context;
	XrmOptionDescList opts;
#endif

// X11 text stuff
#ifdef XLIB
	int		 cw, ch;
#else
	char		 cpustring[15*LINE_SIZE];
#endif
	void		 text(int x, int y, const char *str,
						int rev = 0);
// exit application:
protected:
	char		 exit_flag;
	
// gui update:

protected:
	void		 update_all(void);

protected:
	Widget		create_main_view(int argc, char *argv[]);
	void		create_message_dialog(Widget parent);
	void		create_about_dialog(Widget parent);
	void		create_help_view(Widget parent);
	void		create_author_dialog(Widget parent);
	void		create_cpu_view(Widget parent);
	void		create_bp_dialog(Widget parent);
	void		manage_widget(Widget w);
	String		get_man_page(char *pman_page_name);
	void		initialize_after_create(Widget w, int inverse, char *color);
	void		initialize_after_open(Widget w);
	void 		process_bp_input(KeySym keysym, int which);
	void 		process_button_input(XEvent *pevent);
	void		event_queue(void);

// Internal registers

protected:

	Mc6809* cpu;			// pointer to cpu to send interrupts
	Memory* memory;			// pointer to memory for video update
	Inout* io;			// pointer to io-class handling in/output
	E2video* e2video;		// pointer to video control registers

// Initialisation functions

protected:

	void	reset(void);
	void	initialize_e2window(int argc, char *argv[],
			int inverse, char *color);
	void	initialize_cpu_window(void);

// public interface
public:

	void	set_exit(void);
	void	update(void);
	void	update_1_second(unsigned int state);
	void 	set_new_state(unsigned int new_state);
	void 	update_nafs(void);
#ifndef XLIB
	void 	popup_message(char *pmessage);
	void 	popup_about(void);
	void 	popup_help(void);
	void 	popup_author(Widget w);
	void 	popup_cpu(void);
	void 	popup_bp(unsigned int which);
	void 	popdown_message(void);
	void 	popdown_about(void);
	void 	popdown_help(void);
	void 	popdown_author(Widget w);
	void 	popdown_cpu(void);
	void 	popdown_bp(void);
	void	clear_bp(void);
	void	display_man_page(Widget w);
#endif

protected:
	int	translate_to_ascii(XKeyEvent *pevent);

protected:
	int	oldX, oldY;

#ifndef XLIB
// accelerators:
public:
	void	c_expose(XEvent *pevent);
	void	a_expose(Widget w, XEvent *pevent);
	void	c_keyPress(XEvent *pevent);
#endif

// Public constructor and destructor
public:
		 Gui(
		 	Mc6809* x_cpu,
		 	Memory* x_memory,
		 	Inout* x_io,
		 	E2video* x_video,
		 	int argc,
		 	char *argv[],
		 	char *x_color,
		 	int inverse);
		~Gui();

private:

	void 	update_block(int block_number);

};

extern Gui* ggui;

#endif // __gui_h__

