/*
    inout.cc


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/


#include "inout.h"


Inout::Inout(void)
{
	reset();
} 

Inout::~Inout(void)
{
}


void Inout::reset()
{
	in	         = 0;
	out	         = 0;
	deltaX	         = 0;
	deltaY	         = 0;
	buttonMask       = 0;
	bell_request     = 0;
}


void Inout::put_ch(int key)
{
	if (!key_buffer_full()) {
		key_buffer[in++] = key;
		if (in >= KEY_BUFFER_SIZE)
			in = 0;
	}  // if
} // put_ch

int Inout::key_buffer_full(void)
{
	return ((in == out - 1) || (out == 0 && (in == KEY_BUFFER_SIZE-1)));
}

int Inout::poll(void)
{
	if (in != out)
		return 1;
	else
		return 0;	
}

// input should always be polled before read_ch

Byte Inout::read_ch(void)
{
	int temp;

	while (in == out)    // check for empty buffer
		return 0x00; // should never happen, return a dummy value
	temp = key_buffer[out++];
	if (out >= KEY_BUFFER_SIZE)
		out = 0;
	return temp;
}


void Inout::set_bell(int x_percent)
{
	percent = x_percent;
	bell_request = 1;
}

int Inout::is_bell_request(int *x_percent)
{
	int temp;

	*x_percent = percent;
	temp = bell_request;
	bell_request = 0;		// reset bell request
	return temp;
}

