/*
    main.cc


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/



#include <stdio.h>
#include <signal.h>
#include <limits.h>
#include <time.h>
#include "e2global.h"
#include "memory.h"
#include "pia1.h"
#include "pia2.h"
#include "e2floppy.h"
#include "e2video.h"
#include "mc146818.h"
#include "da6809.h"
#include "mc6809.h"
#include "inout.h"

#ifdef __osf__
extern "C" unsigned int alarm(unsigned int);
#endif

#ifdef sun
extern int optind;
extern int opterr;
extern char *optarg;
#endif

#ifdef ultrix
extern int optind;
extern int opterr;
extern char *optarg;
#endif


// global variables
Mc6809*	gcpu;
Gui*	ggui;
Memory* memory;


#ifdef SIGALRM
#ifdef sun
void update(int, ...)
#else
void update(int)
#endif
{
	if (gcpu)
		gcpu->update_flag = 1;
	(void)signal(SIGALRM, update);
	alarm(1);
}
#endif // SIGALRM


int main(int argc, char *argv[])
{
	char *drives[3];
	char *hex_file;
	char *color;
	int  inverse;
	int  result;
	Inout* io;
	Da6809* disassembler;

	drives[0] = "system";
	drives[1] = "work";
	drives[2] = "";
	drives[3] = "";
	hex_file  = "neumon54.hex";
	color     = "green";
	inverse   = 0;

#ifdef linux
	optind = 0;
#endif
	opterr = 0;
	while (1) {
		result = getopt(argc, argv, "ihf:0:1:2:3:c:");
		if (result == -1)
			break;
		switch(result) {
			case 'f': hex_file  = optarg; break;
			case '0': drives[0] = optarg; break;		
			case '1': drives[1] = optarg; break;
			case '2': drives[2] = optarg; break;
			case '3': drives[3] = optarg; break;
			case 'c': color     = optarg; break;
			case 'i': inverse   = 1; break;
			case 'h':
				fprintf(stderr, "usage: flexemu <options>\n");
				fprintf(stderr, "  <options> are:\n");
				fprintf(stderr, "  -f <hexfile>\n");
				fprintf(stderr, "  -0 <diskimage/directory drive 0>\n");
				fprintf(stderr, "  -1 <diskimage/directory drive 1>\n");
				fprintf(stderr, "  -2 <diskimage/directory drive 2>\n");
				fprintf(stderr, "  -3 <diskimage/directory drive 3>\n");
				fprintf(stderr, "  -c <color>\n");
				fprintf(stderr, "  -i (display inverse video)\n");
				fprintf(stderr, "  -h (display this)\n");
				exit(EXIT_SUCCESS);
		}  // switch
	} // while


	memory		  = new Memory;
	disassembler	  = new Da6809(memory);
	gcpu	          = new Mc6809(memory, disassembler);
	io		  = new Inout();
	memory->video     = new E2video();
	ggui 		  = new Gui(gcpu, memory, io, memory->video,
				argc, argv, color, inverse);
	gcpu->set_gui(ggui);
	memory->pia1      = new Pia1(io, gcpu);
	memory->pia2      = new Pia2(io, gcpu);
	memory->fdc       = new E2floppy(gcpu, drives);
	memory->rtc       = new Mc146818(gcpu);

//	(void)signal(SIGINT, SIG_IGN);
#ifdef SIGALRM
	(void)signal(SIGALRM, update);
	alarm(1);
#endif

	gcpu->load_hexfile(hex_file);

	memory->reset();
	gcpu->reset();
	gcpu->statemachine();

	delete memory->rtc;
	delete memory->fdc;
	delete memory->pia2;
	delete memory->pia1;
	delete ggui;
	delete memory->video;
	delete io;
	delete gcpu;
	delete disassembler;
	delete memory;

	exit(0);
}
