/*
    memory.h


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/



#ifndef __memory_h__
#define __memory_h__

#include "e2global.h"
#include "misc.h"
#include "e2.h"

#define READ_DEFAULT	0
#define FDC_READSTATUS	1
#define FDC_READ	2
#define PIA1_READ	3
#define PIA2_READ	4
#define RTC_READ	5
#define READ_UNSELECTED	6

#define NO_WRITE	0
#define FDC_WRITE	1
#define FDC_WRITEDRISEL	2
#define WRITE_MMU	3
#define PIA1_WRITE	4
#define PIA2_WRITE	5
#define VIDEO_WRITE_CO	6
#define RTC_WRITE	7


class Memory {

// Internal registers

private:

	Byte *ppage[16];
	Byte memory[0x10000L];
	Byte byte_conversion_table[256];
	Byte read_tab[(~GENIO_MASK & 0xffff) + 1];
	Byte write_tab[(~GENIO_MASK & 0xffff) + 1];

public:
	// interface to video display
	Byte video_ram[VIDEORAM_SIZE * 12];
	char changed[YBLOCKS];

	// interface to memory mapped I/O
	E2video*	video;
//	Uart1*		uart1;
	Pia1*		pia1;
	Pia2*		pia2;
	E2floppy*	fdc;
	Mc146818*	rtc;

// Initialisation functions

public:

	void	initialize_byte_conv_table(int lsb_first);

// public interface
public:

	void	init(void);
	void	reset(void);
	Byte	read(Word addr);
	Word	read_word(Word addr);
	Byte	read_unselected(Word addr);
	void	write(Word addr, Byte val);
	void	write_word(Word addr, Word val);
	void	write_mmu(Word addr, Byte val);
	void	write_rom(Word addr, Byte val);

// Public constructor and destructor
public:
		 Memory();
		~Memory();
};

#endif // __memory_h__

