/*
    nafs.cc


    flexemu, an MC6809 emulator running FLEX
    Copyright (C) 1997  W. Schwotzer

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    woschwot@si.bosch.de
*/



// support of the native file system

#ifdef NAFS
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <time.h>
#include <utime.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "misc.h"
#include "e2floppy.h"


int E2floppy::update(void)
{
	struct s_floppy *pfloppy;
	int i;

	for (i = 0; i < 4; i++)
		if (open_files(i))
			return 0;
	for (i = 0; i < 4; i++) {
		pfloppy = &floppy[i];
		if (pfloppy->type == NAFS_FORMAT) {
			umount_nafs(i);
			mount_nafs(pfloppy, i);
		} // if
	}
	return 1;
} //update


void E2floppy::header_for_nafs_format(s_floppy *pfloppy, int wp) {
	pfloppy->ready         = 1;
	pfloppy->type          = NAFS_FORMAT;
	pfloppy->offset        = 0;
	pfloppy->fp            = NULL;
	pfloppy->write_protect = wp;
	pfloppy->max_sector    = MAX_SECTOR >> 1;
	pfloppy->max_sector0   = (INIT_DIR_SECTORS + 4) >> 1;
	pfloppy->max_track     = MAX_TRACK;
	pfloppy->byte_p_sector = 256;
	pfloppy->byte_p_track0 = pfloppy->max_sector0 << 9; 
	pfloppy->byte_p_track  = pfloppy->max_sector << 9; 
	pfloppy->pflex_links     = (struct s_link_table *)calloc(LINK_TABLE_SIZE, sizeof(struct s_link_table));
	pfloppy->pflex_sys_info  = (struct s_sys_info_sector *)calloc(2, sizeof(s_sys_info_sector));
	pfloppy->pflex_unused    = (struct s_unused_sector *)calloc(1, sizeof(s_unused_sector));
	pfloppy->pflex_directory = (struct s_dir_sector *)calloc(INIT_DIR_SECTORS, sizeof(struct s_dir_sector));
	pfloppy->dir_sectors	 = INIT_DIR_SECTORS;
	pfloppy->pnew_file       = (struct s_new_file *)calloc(INIT_NEW_FILES, sizeof(s_new_file));
	pfloppy->new_files	 = INIT_NEW_FILES;
	pfloppy->dir_extend.sec_trk = 0;
	if (pfloppy->pflex_links == NULL || pfloppy->pflex_sys_info == NULL ||
	    pfloppy->pflex_unused == NULL || pfloppy->pflex_directory == NULL ||
	    pfloppy->pnew_file == NULL) {
	    	fprintf(stderr, "error allocating memory, can't continue\n");
	    	exit(1);
	} // if
} // header_for_nafs_format

// if valid flex filename return is > 0 otherwise = 0
/*
	some examples:

	allowed:	x.a xx.a xxxxxxxx.a x xx xxxxxxxx
	not allowed:	x. .a xx. xxxxxxxxx.a X.a xxxxxxxxX.a
*/
int E2floppy::is_flex_filename(char *pfilename, char *pname, char *pext)
{
	int result;
	char dot;

	*pname = '\0';
	*pext  = '\0';
	dot    = '\0';
	result = sscanf(pfilename, "%1[a-z]%7[a-z0-9_-]", pname, (pname + 1));
	if (!result || result == EOF)
		return 0;
	if (result == 1)
		result = sscanf(pfilename, "%*1[a-z]%c%1[a-z]%2[a-z0-9_-]",
		&dot, pext, (pext + 1));
	else
		result = sscanf(pfilename, "%*1[a-z]%*7[a-z0-9_-]%c%1[a-z]%2[a-z0-9_-]",
		&dot, pext, (pext + 1));
	if (!result || result == 1 || result == EOF)
		return 0;
	if (strlen(pname) + strlen(pext) + (dot == '.' ? 1 : 0) != strlen(pfilename))
		return 0;
	strupper(pname);
	strupper(pext);
	return 1;
} // is_flex_filename


void E2floppy::initialize_flex_directory(struct s_floppy *pfloppy)
{
	int i, j;
	char *p;

	for (i = 0; i < pfloppy->dir_sectors; i++) {
		p = (char *)(pfloppy->pflex_directory + i);
		for (j=0; j < 256; j++)
			*(p++) = 0x00;
		if (i < pfloppy->dir_sectors - 1) {
			(pfloppy->pflex_directory+i)->next_trk = (i+5) / MAX_SECTOR;
			(pfloppy->pflex_directory+i)->next_sec =
				((i+5) % MAX_SECTOR) + 1;
		} // if
	} // for
} // initialize_flex_directory


void E2floppy::scan_unix_filename(char *pfn, char *pfilename)
{
	char *p;
	int i, j;

	j = 0;
	p = (char *)pfn;
	if (*p != -1) {
		// get file name
		for (i=0; i < 8 && *p; i++) {
			*(pfilename+i) = tolower(*p);
			p++;
		}
		if (i > 0) {
			p += 8 - i;
			*(pfilename + (i++)) = '.';
			// get file extension
			for (j=i; j < i+3 && *p; j++) {
				*(pfilename+j) = tolower(*p);
				p++;
			} // for
		} // if
	} // if
	*(pfilename+j) = '\0';
} // scan_unix_filename


// return unix filename from a pointer to a flex directory entry 
char *E2floppy::to_unix_filename(s_dir_entry *pde)
{
	static char filename[12];

	scan_unix_filename((char *)pde, (char *)&filename);
	return (char *)&filename;
} // to_unix_filename


// return unix filename from a given file_id
char *E2floppy::unix_filename(int file_id)
{
	static char filename[12];

	if (file_id < 0)
		sprintf((char *)&filename, "tmp%02d",  NEW_FILE1 - file_id);
	else
		scan_unix_filename(
		   (char *)&((pfs->pflex_directory + (file_id/10))->dir_entry[file_id%10].filename),
		   (char *)&filename);
	return (char *)&filename;
} // unix_filename


int E2floppy::index_of_new_file(Byte track, Byte sector)
{
	int i;
	char path[256];
	s_new_file *pnf;

	i = 0;
	while ((pfs->pnew_file+i)->first.sec_trk && i < pfs->new_files) {
		if (tr == (pfs->pnew_file+i)->next.st.trk &&
		    sr == (pfs->pnew_file+i)->next.st.sec)
			return i;
		i++;
	} // while
	if (i >= pfs->new_files) {
		// new file table is full, increase new file table by 2
		pnf = (s_new_file *)calloc(pfs->new_files + 2,
			sizeof(struct s_new_file));
		if (pnf == NULL)
			return -1; // can't allocate memory
		memcpy(pnf, pfs->pnew_file,
			pfs->new_files * sizeof(struct s_new_file));
		free(pfs->pnew_file);
		pfs->pnew_file = pnf;
		pfs->new_files += 2;
		(pfs->pnew_file+(pfs->new_files-2))->first.sec_trk = 0;
		(pfs->pnew_file+(pfs->new_files-1))->first.sec_trk = 0;
	} // if
	strcpy((char *)&(pfs->pnew_file+i)->filename,
		unix_filename(NEW_FILE1 - i));
	strcpy((char *)&path, pfs->ppath);
	strcat((char *)&path, "/");
	strcat((char *)&path, (pfs->pnew_file+i)->filename);
#ifdef sun
	(pfs->pnew_file+i)->fp = fopen(path, "w");
#else
	(pfs->pnew_file+i)->fp = fopen(path, "wb");
#endif
	if ((pfs->pnew_file+i)->fp == NULL)
		return -1;
	(pfs->pnew_file+i)->first.st.trk = track;
	(pfs->pnew_file+i)->first.st.sec = sector;
	(pfs->pnew_file+i)->next.sec_trk = 0;
	(pfs->pnew_file+i)->f_record     = 0;
	return i;
} // index_of_new_file


int E2floppy::extend_directory(struct s_floppy *pfloppy,
	int index, s_dir_sector *pdb)
{
	struct s_dir_sector *pfd;

	// increase flex_directory by one sector
	pfd = (s_dir_sector *)calloc(pfloppy->dir_sectors + 1,
		sizeof(struct s_dir_sector));
	if (pfd == NULL)
		return -1; // can't allocate memory
	memcpy(pfd, pfloppy->pflex_directory,
		pfloppy->dir_sectors * sizeof(struct s_dir_sector));
	memcpy(pfd+pfloppy->dir_sectors, pdb, 256);
	(pfloppy->pflex_links+index)->f_record = pfloppy->dir_sectors;
	(pfloppy->pflex_links+index)->file_id  = DIRECTORY;
	free(pfloppy->pflex_directory);
	pfloppy->pflex_directory = pfd;
	pfloppy->dir_sectors++;
	pfloppy->dir_extend.sec_trk = 0;// reset directory extend track/sector
	return 0;
} // extend_directory


t_st *E2floppy::link_address(void)
{
	static t_st link;
	s_dir_entry *pd;
	int i;

	link.sec_trk = 0;
	for (i = 0; i < pfs->dir_sectors * 10; i++) {
		pd = &((pfs->pflex_directory + (i / 10))->dir_entry[i % 10]);
		if (!strncmp(pd->filename, "FLEX\0\0\0\0", 8) &&
			!strncmp(pd->file_ext, "SYS", 3)) {
			link.st.trk = pd->start_trk;		
			link.st.sec = pd->start_sec;
			break;
		}		
	} // for
	return &link;
}  // link_address


//if directory can't be extended return -1
int E2floppy::next_free_dir_entry(struct s_floppy *pfloppy)
{
	int i, j, index;
	s_sys_info_sector *psis;
	char sector_buffer[256];
	int trk, sec;

	for (i = 0; i < pfloppy->dir_sectors * 10; i++) {
		if (!((pfloppy->pflex_directory + (i / 10))->dir_entry[i % 10].filename[0]))
			return i;
	} // for
	psis = pfloppy->pflex_sys_info;
	for (j = 0; j < 256; j++)
		sector_buffer[j] = 0x00;
	sector_buffer[3] = pfloppy->dir_sectors - INIT_DIR_SECTORS + 1;
	trk = psis->fc_start_trk;
	sec = psis->fc_start_sec;
	index = trk * MAX_SECTOR + sec - 1;
	if (!extend_directory(pfloppy, index, (s_dir_sector *)sector_buffer)) {
		(pfloppy->pflex_directory+pfloppy->dir_sectors-2)->next_trk = trk;
		(pfloppy->pflex_directory+pfloppy->dir_sectors-2)->next_sec = sec;
		if (++psis->fc_start_sec > MAX_SECTOR) {
			psis->fc_start_sec = 1;
			psis->fc_start_trk++;
		}
		if (--psis->free[1] == 0xff)
			psis->free[0]--;
		return i;
	} else
		return -1;
}  // next_free_dir_entry


void E2floppy::initialize_flex_link_table(struct s_floppy *pfloppy)
{
	unsigned int i, fc_start, free;
	struct s_link_table *plink;
	struct s_sys_info_sector *psis;

	fc_start = MAX_SECTOR;
	// entries for system and directory sectors
	for (i = 0; i < fc_start; i++) {
		plink = pfloppy->pflex_links + i;
		plink->next.sec_trk	= 0;
		plink->record_nr[0]	= 0;
		plink->record_nr[1]	= 0;
		plink->f_record		= i < 4 ? 0 : i - 4;
		plink->file_id		= i < 4 ? SYSTEM : DIRECTORY;
	} // for
	for (i = fc_start; i < LINK_TABLE_SIZE; i++) {
		plink = pfloppy->pflex_links + i;
		if (i == LINK_TABLE_SIZE-1)
			plink->next.sec_trk = 0;
		else {
			plink->next.st.trk = (i+1) / MAX_SECTOR;
			plink->next.st.sec = ((i+1) % MAX_SECTOR) + 1;
		} // else
		plink->record_nr[0]	= 0;
		plink->record_nr[1]	= 0;
		plink->f_record		= 0;
		plink->file_id		= FREE_CHAIN;
	} // for
	free = LINK_TABLE_SIZE - fc_start;
	// and now update system info sectors
	for (i = 0; i < 2; i++) {
		psis = pfloppy->pflex_sys_info + i;
		psis->fc_start_trk	= fc_start / MAX_SECTOR;
		psis->fc_start_sec	= (fc_start % MAX_SECTOR) + 1;
		psis->fc_end_trk	= MAX_TRACK;
		psis->fc_end_sec	= MAX_SECTOR;
		psis->free[0]		= free >> 8;
		psis->free[1] 		= free & 0xff;
	} // for
} // initialize_flex_link_table


void E2floppy::initialize_new_file_table(struct s_floppy *pfloppy)
{
	int i;

	for (i = 0; i < pfloppy->new_files; i++) {
		(pfloppy->pnew_file+i)->first.sec_trk = 0;
		(pfloppy->pnew_file+i)->next.sec_trk = 0;
	} // for
} // initialize_new_file_table

// check for any open file in an nafs file system
int E2floppy::open_files(int index)
{
	struct s_floppy *pfloppy;
	int i;

	pfloppy = &floppy[index];
	if (pfloppy->type == NAFS_FORMAT) {
		for (i = 0; i < pfloppy->new_files; i++) {
			if ((pfloppy->pnew_file+i)->first.sec_trk != 0)
				return 1;
		} // for
	} // if
	return 0;
} // open_files


void E2floppy::close_new_files(int index)
{
	struct s_floppy *pfloppy;
	int i;
	int first = 1;

	pfloppy = &floppy[index];
	if (pfloppy->type == NAFS_FORMAT) {
		for (i = 0; i < pfloppy->new_files; i++) {
			if ((pfloppy->pnew_file+i)->first.sec_trk != 0) {
				if (first) { 
					fprintf(stderr, "flexemu warning: there are still open files temporarily stored as:\n");
					first = 0;
				}
				fclose((pfloppy->pnew_file+i)->fp);
				fprintf(stderr, "   %s\n", (pfloppy->pnew_file+i)->filename);
			} // if
		} // for
	} // if
} // close_new_files


void E2floppy::free_nafs_memory(int index)
{
	int i;

	if ((floppy[index].type == NAFS_FORMAT)) {
		free(floppy[index].pflex_links);
		free(floppy[index].pflex_directory);
		free(floppy[index].pflex_sys_info);
		free(floppy[index].pflex_unused);
		free(floppy[index].pnew_file);
	} // if
} // free_nafs_memory

// if file won't fit return 0 otherwise return 1
int E2floppy::add_to_link_table(struct s_floppy *pfloppy, int dir_index,
	off_t size, int random, t_st *pbegin, t_st *pend)
{
	unsigned int i, free, begin, records;
	struct s_link_table *plink;
	struct s_sys_info_sector *psis;

	psis = pfloppy->pflex_sys_info;	
	free = (psis->free[0] << 8) + psis->free[1];
	if (size > (off_t)(free * 252))
		return 0;
	records = (size + 251) / 252;
	pbegin->st.sec = psis->fc_start_sec;
	pbegin->st.trk = psis->fc_start_trk;
	begin = (pbegin->st.trk * MAX_SECTOR) + pbegin->st.sec - 1;
	for (i = 1; i <= records; i++) {
		plink = pfloppy->pflex_links + i + begin - 1;
		if (i == records) 
			plink->next.sec_trk = 0;
		if (random) {
			plink->record_nr[0]	= i > 2 ? (i-2) >> 8 : 0;
			plink->record_nr[1]	= i > 2 ? (i-2) & 0xff : 0;
		} else {
			plink->record_nr[0]	= i >> 8;
			plink->record_nr[1]	= i & 0xff;
		}
		plink->f_record		= i-1;
		plink->file_id		= dir_index;
	} // for
	pend->st.sec = ((i + begin - 2) % MAX_SECTOR) + 1;
	pend->st.trk = (i + begin - 2) / MAX_SECTOR;
	// update sys info sector
	psis->fc_start_sec = ((i + begin - 1) % MAX_SECTOR) + 1;
	psis->fc_start_trk = (i + begin -1) / MAX_SECTOR;
	psis->free[0] = (free - records) >> 8;
	psis->free[1] = (free - records) & 0xff;
	return 1;
} // add_to_link_table


void E2floppy::add_to_directory(struct s_floppy *pfloppy, char *name,
	char *ext, int dir_index, int random, struct stat *pstat,
	t_st *pbegin, t_st *pend, int wp)
{
	s_dir_entry *pd;
	struct tm   *lt;
	unsigned int records;

	lt = localtime(&(pstat->st_mtime));
	records = (pstat->st_size + 251) / 252;
	pd = &((pfloppy->pflex_directory + (dir_index / 10))->dir_entry[dir_index % 10]);
	strcpy(pd->filename, name);
	strcpy(pd->file_ext, ext);
	pd->file_attr	= wp ? 0xc0 : 0x00;
	pd->start_trk	= pbegin->st.trk;
	pd->start_sec	= pbegin->st.sec;
	pd->end_trk	= pend->st.trk;
	pd->end_sec	= pend->st.sec;
	pd->records[0]	= records >> 8;
	pd->records[1]	= records & 0xff;
	pd->sector_map	= random ? 0x02 : 0x00;
	pd->month	= lt->tm_mon + 1;
	pd->day		= lt->tm_mday;
	pd->year	= lt->tm_year;
} // add_to_directory


void E2floppy::modify_random_file(char *path, struct stat *pstat, t_st begin)
{
	char file_sector_map[252 * 2];
	int i, n, data_size, index;
	FILE *fp;

	data_size = pstat->st_size - (252 * 2);
	if (data_size >= 252 * 2) {
		index = begin.st.trk * MAX_SECTOR + begin.st.sec - 1 + 2;
		for (i = 0; i < 252 * 2; i++)
			file_sector_map[i] = 0;
		for (n = 0; n < (data_size / (252 * 255)) ; n++) {
			file_sector_map[3*n]   = index / MAX_SECTOR;
			file_sector_map[3*n+1] = index % MAX_SECTOR + 1;
			file_sector_map[3*n+2] = 255;
			index += 255;
		} // for
		i = data_size % (252 * 255);
		if (i) {
			file_sector_map[3*n]   = index / MAX_SECTOR;
			file_sector_map[3*n+1] = index % MAX_SECTOR + 1;
			file_sector_map[3*n+2] = (i+251) / 252;
		} // if
#ifdef sun
		fp = fopen(path, "r+");
#else
		fp = fopen(path, "rb+");
#endif
		if (fp != NULL)
			fwrite(&file_sector_map, 252, 2, fp);
		fclose(fp);
	} // if
} // modify_random_file

// dwp = write protect for drive

void E2floppy::fill_flex_directory(struct s_floppy *pfloppy, int dwp)
{
	DIR *pd;
	struct dirent *pentry;
	char name[9], ext[4], path[256], *pfilename;
	int dir_index, random;
	t_st begin, end;
	struct stat sbuf;

	initialize_flex_directory(pfloppy);
	initialize_flex_link_table(pfloppy);
	if ((pd = opendir(pfloppy->ppath)) != NULL) {
		dir_index = 0;
		while ((pentry = readdir(pd)) != NULL) {
			pfilename = (char *)&pentry->d_name;
			strcpy((char *)&path, pfloppy->ppath);
			strcat((char *)&path, "/");
			strcat((char *)&path, pfilename);
			if (is_flex_filename(pfilename, (char *)&name, (char *)&ext) &&
			   !stat(path, &sbuf) &&
			   sbuf.st_size > 0 &&
			   (dir_index = next_free_dir_entry(pfloppy)) >= 0) {
			   	random = sbuf.st_mode & S_IXUSR;
			   	if (add_to_link_table(pfloppy, dir_index,
			   		sbuf.st_size, random, &begin, &end)) {
			   		add_to_directory(pfloppy, name, ext,
			   	 	   dir_index, random, &sbuf, &begin, &end,
			   		   dwp || access(path, W_OK));
			   		if (random)
			   			modify_random_file(path,
			   				&sbuf, begin);
				}
			}
		} // while
		closedir(pd);
	} //if 
} // initialize_flex_directory


void E2floppy::initialize_flex_sys_info_sectors(struct s_floppy *pfloppy,
	int number)
{
	unsigned int i;
	char name[9];
	char ext[4];
	struct stat sbuf;
	struct tm *lt;
	struct s_sys_info_sector *psis;

	if(!stat(pfloppy->ppath, &sbuf)) {
		psis = pfloppy->pflex_sys_info;
		lt = localtime(&(sbuf.st_mtime));
		name[0] = '\0';
		ext[0]  = '\0';
		sscanf(pfloppy->ppath, "%8[a-zA-Z0-9_-].%3[a-zA-Z0-9_-]",
			name, ext);
		if (!name[0])
			strcpy(name, "UNIX");
		else {
			strupper(name);
			strupper(ext);
		} // else
		for (i=0; i < 16; i++)
			psis->unused1[i] = 0;
		strncpy((char *)&(psis->disk_name), name, 8);
		strncpy((char *)&(psis->disk_ext), ext, 3);
		psis->disk_number[0] = number >> 8;
		psis->disk_number[1] = number & 0xff;
		psis->fc_start_trk   = 0;
		psis->fc_start_sec   = 0;
		psis->fc_end_trk     = 0;
		psis->fc_end_sec     = 0;
		psis->free[0]        = 0;
		psis->free[1]        = 0;
		psis->month          = lt->tm_mon + 1;
		psis->day            = lt->tm_mday;
		psis->year           = lt->tm_year;
		psis->last_trk       = MAX_TRACK;
		psis->last_sec       = MAX_SECTOR;
	
		for (i=0; i < 216; i++)
			psis->unused2[i] = 0;
		memcpy(pfloppy->pflex_sys_info + 1, pfloppy->pflex_sys_info,
			sizeof(struct s_sys_info_sector));
	} // if
} // initialize_flex_sys_info_sectors


// I don't know what this sector should be used for but
// emulate it anyway
void E2floppy::initialize_flex_unused_sector(struct s_floppy *pfloppy)
{
	int i;

	pfloppy->pflex_unused->next_trk = 0x00;
	pfloppy->pflex_unused->next_sec = 0x03;
	for (i=0; i < 254; i++)
		pfloppy->pflex_unused->unused[i] = 0;
} // initialize_flex_unused_sector


// change the file id in flex_links
void E2floppy::change_file_id(int index, int old_id, int new_id)
{
	int i;

	i = index;
	while (i >= 0 && (pfs->pflex_links+i)->file_id == old_id) {
		(pfs->pflex_links+i)->file_id = new_id;
		i = (pfs->pflex_links+i)->next.st.trk * MAX_SECTOR +
			(pfs->pflex_links+i)->next.st.sec - 1;
	} // while
} // change_file_id


void E2floppy::check_for_delete(int dir_index, s_dir_sector *pdb)
{
	int i, index;
	char path[256], *pfilename;
	s_dir_sector *pd;

	pd = pfs->pflex_directory + dir_index;
	for (i = 0; i < 10; i++) {
		if (pdb->dir_entry[i].filename[0] == -1 &&
		   pd->dir_entry[i].filename[0] != -1) {
			pfilename = unix_filename(dir_index * 10 + i);
			index = pd->dir_entry[i].start_trk * MAX_SECTOR +
				pd->dir_entry[i].start_sec - 1;
			strcpy((char *)&path, pfs->ppath);
			strcat((char *)&path, "/");
			strcat((char *)&path, pfilename);
			unlink(path);
			change_file_id(index, dir_index * 10 + i, FREE_CHAIN);
			break;
		} // if
	} // for	

} // check_for_delete


void E2floppy::check_for_rename(int dir_index, s_dir_sector *pdb)
{
	int i;
	char old_path[256], new_path[256];
	char *pfilename1, *pfilename2;

	for (i = 0; i < 10; i++) {
		pfilename1 = unix_filename(dir_index * 10 + i);
		pfilename2 = to_unix_filename(&pdb->dir_entry[i]);
		if (*pfilename1 && *pfilename2 &&
		   strcmp(pfilename1, pfilename2) != 0) {
			strcpy((char *)&old_path, pfs->ppath);
			strcat((char *)&old_path, "/");
			strcat((char *)&old_path, pfilename1);
			strcpy((char *)&new_path, pfs->ppath);
			strcat((char *)&new_path, "/");
			strcat((char *)&new_path, pfilename2);
			rename(old_path, new_path);
			break;
		} // if
	} // for	
} // check_for_rename


void E2floppy::check_for_extend(int dir_index, s_dir_sector *pdb)
{
	s_dir_sector *pd;

	pd = pfs->pflex_directory + dir_index;
	if (!pd->next_sec && !pd->next_trk && (pdb->next_trk || pdb->next_sec)) {
		   pfs->dir_extend.st.trk = pdb->next_trk;
		   pfs->dir_extend.st.sec = pdb->next_sec;
	} // if
} // check_for_extend


// return -1 if not successful otherwise return 0
// years < 70 will be represented as >= 2000
int E2floppy::set_file_time(char *ppath, Byte month, Byte day, Byte year)
{
	struct stat    statbuf;
	struct utimbuf timebuf;
	struct tm      file_time;

	if (stat(ppath, &statbuf) >= 0) {
		timebuf.actime = statbuf.st_atime;
		file_time.tm_sec   = 0;
		file_time.tm_min   = 0;
		file_time.tm_hour  = 0;
		file_time.tm_mon   = month-1;
		file_time.tm_mday  = day;
		file_time.tm_year  = year < 70 ? year + 100 : year;
		file_time.tm_isdst = 0;
		timebuf.modtime    = mktime(&file_time);
		if (timebuf.modtime >= 0 && utime(ppath, &timebuf) >= 0)
				return 0;
			else
				return -1;
	} // if
	return -1;
} // set_file_time


int E2floppy::check_for_new_file(int dir_index, s_dir_sector *pd)
{
	int i, j, index;
	char old_path[256], new_path[256];

	j = 0;
	while ((pfs->pnew_file+j)->first.sec_trk && j < pfs->new_files) {
		for (i = 0; i < 10; i++) {
			if ((pfs->pnew_file+j)->first.st.sec == pd->dir_entry[i].start_sec &&
			    (pfs->pnew_file+j)->first.st.trk == pd->dir_entry[i].start_trk) {
				index = (pfs->pnew_file+j)->first.st.trk * MAX_SECTOR +
			   	   (pfs->pnew_file+j)->first.st.sec - 1; 
				change_file_id(index, NEW_FILE1-j,
					10 * dir_index + i);
				fclose((pfs->pnew_file+j)->fp);
				strcpy((char *)&old_path, pfs->ppath);
				strcat((char *)&old_path, "/");
				strcat((char *)&old_path, (pfs->pnew_file+j)->filename);
				strcpy((char *)&new_path, pfs->ppath);
				strcat((char *)&new_path, "/");
				strcat((char *)&new_path,
					unix_filename((pfs->pflex_links+index)->file_id));
				rename(old_path, new_path); 
				set_file_time(new_path, pd->dir_entry[i].month,
				   pd->dir_entry[i].day, pd->dir_entry[i].year);
				// remove entry in new file table and
				// move following entries
				while ((pfs->pnew_file+j+1)->first.sec_trk &&
					j < pfs->new_files-1) {
					strcpy((char *)&(pfs->pnew_file+j)->filename,
						(pfs->pnew_file+j+1)->filename);
					(pfs->pnew_file+j)->first.sec_trk = (pfs->pnew_file+j+1)->first.sec_trk;
					(pfs->pnew_file+j)->next.sec_trk  = (pfs->pnew_file+j+1)->next.sec_trk;
					(pfs->pnew_file+j)->f_record      = (pfs->pnew_file+j+1)->f_record;
					(pfs->pnew_file+j)->fp            = (pfs->pnew_file+j+1)->fp;
					index = (pfs->pnew_file+j)->first.st.trk * MAX_SECTOR +
			   		   (pfs->pnew_file+j)->first.st.sec - 1; 
					change_file_id(index, NEW_FILE1-j-1, NEW_FILE1-j);
					j++;
				} // while
				(pfs->pnew_file+j)->first.sec_trk = 0;
				return 0;
			} // if
		} // for
		j++;
	} // while
	return 0;
} // check_for_new_file


int E2floppy::last_of_free_chain(Byte tr, Byte sr)
{
	return (pfs->pflex_sys_info->fc_end_trk == tr &&	
		pfs->pflex_sys_info->fc_end_sec == sr);
} // last_of_free_chain


void E2floppy::read_nafs_sector(void)
{
	char	*p, path[256], *mode;
	int	 i, di, bytes;
	FILE	*fp;
	t_st	*link;

#ifdef sun
	mode = "r";
#else
	mode = "rb";
#endif
	bytes = 0;	// satisfy compiler
	i = tr * MAX_SECTOR + (sr - 1);
	switch ((pfs->pflex_links+i)->file_id) {
		case SYSTEM :
			p = NULL; // satisfy compiler
			if (sr >= 3)
				p = (char *)(pfs->pflex_sys_info + sr - 3);
			else if (sr == 2)
				p = (char *)pfs->pflex_unused;
			else if (sr == 1) {
				link = link_address();
				strcpy((char *)&path, pfs->ppath);
				strcat((char *)&path, "/boot");
				if ((fp = fopen(path, mode)) != NULL) {
					bytes = fread(&sector_buffer, 1, 256, fp);
					fclose(fp);
				}
				if (bytes != 256) {
					drq = 0;
					str = 0x10;
					byteCount = 0;
				}
				sector_buffer[3] = link->st.trk;
				sector_buffer[4] = link->st.sec;
				break;
			} else
				break;
			memcpy(sector_buffer, p, pfs->byte_p_sector);
			break;
		case DIRECTORY:
			di = (pfs->pflex_links+i)->f_record;
			p = (char *)(pfs->pflex_directory + di);
			memcpy(&sector_buffer, p, pfs->byte_p_sector);
			break;
		case FREE_CHAIN :
			sector_buffer[0] = (pfs->pflex_links+i)->next.st.trk;
			sector_buffer[1] = (pfs->pflex_links+i)->next.st.sec;
			sector_buffer[2] = (pfs->pflex_links+i)->record_nr[0];
			sector_buffer[3] = (pfs->pflex_links+i)->record_nr[1];
			// free chain sector reads always
			// filled with zero
			for (i = 4; i < 256; i++)
				sector_buffer[i] = 0;
			break;
		default : 
			strcpy((char *)&path, pfs->ppath);
			strcat((char *)&path, "/");
			strcat((char *)&path,
				unix_filename((pfs->pflex_links+i)->file_id));
			sector_buffer[0] = (pfs->pflex_links+i)->next.st.trk;
			sector_buffer[1] = (pfs->pflex_links+i)->next.st.sec;
			sector_buffer[2] = (pfs->pflex_links+i)->record_nr[0];
			sector_buffer[3] = (pfs->pflex_links+i)->record_nr[1];
			if ((fp = fopen(path, mode)) != NULL &&
				!fseek(fp, (pfs->pflex_links+i)->f_record * 252L, SEEK_SET)) {
				bytes = fread(&sector_buffer[4], 1, 252, fp);
				fclose(fp);
			// stuff last sector of file with 0
				for (i = 4 + bytes; i < 256; i++)
					sector_buffer[i] = 0;
			} else { // unable to read sector
				drq = 0;
				str = 0x10;
				byteCount = 0;
			} // else
	} // switch
} //read_nafs_sector


void E2floppy::write_nafs_sector(void)
{
	char	*p, path[256], *mode;
	int	 i, di, bytes, new_file_index;
	FILE	*fp;

#ifdef sun
	mode = "r+";
#else
	mode = "rb+";
#endif
	bytes = 0;	// satisfy compiler
	i = tr * MAX_SECTOR + (sr - 1);
	switch ((pfs->pflex_links+i)->file_id) {
		case SYSTEM :
			p = NULL; // satisfy compiler
			if (sr >= 3)
				p = (char *)(pfs->pflex_sys_info + sr - 3);
			else if (sr == 2)
				p = (char *)pfs->pflex_unused;
			else if (sr == 1) {
#ifdef sun
				mode = "w";
#else
				mode = "wb";
#endif
				strcpy((char *)&path, pfs->ppath);
				strcat((char *)&path, "/boot");
				if ((fp = fopen(path, mode)) != NULL) {
					bytes = fwrite(&sector_buffer, 1, 256, fp);
					fclose(fp);
				}
				if (bytes != 256)
					str = 0x10;
				break;
			} else
				break;
			memcpy(p, sector_buffer, pfs->byte_p_sector);
			break;
		case DIRECTORY:
			di = (pfs->pflex_links+i)->f_record;
			p = (char *)(pfs->pflex_directory + di);
			check_for_delete(di, (s_dir_sector *)sector_buffer);
			check_for_new_file(di, (s_dir_sector *)sector_buffer);
			check_for_rename(di, (s_dir_sector *)sector_buffer);
			check_for_extend(di, (s_dir_sector *)sector_buffer);
			memcpy(p, sector_buffer, pfs->byte_p_sector);
			break;
		case FREE_CHAIN :
			if (pfs->dir_extend.st.trk == tr &&
			   pfs->dir_extend.st.sec == sr) {
			   	extend_directory(pfs, i, (s_dir_sector *)sector_buffer);
				break;
			} // if
			if (last_of_free_chain(tr, sr) &&
				(sector_buffer[1] || sector_buffer[0])) {
				(pfs->pflex_links+i)->next.st.trk  = sector_buffer[0];
				(pfs->pflex_links+i)->next.st.sec  = sector_buffer[1];
				(pfs->pflex_links+i)->record_nr[0] = sector_buffer[2];
				(pfs->pflex_links+i)->record_nr[1] = sector_buffer[3];
				break;
			}
			if ((new_file_index = index_of_new_file(tr, sr)) < 0) {
				str = 0x10; // no more new files can be created
				break;
			}
			(pfs->pnew_file+new_file_index)->next.st.trk = sector_buffer[0];
			(pfs->pnew_file+new_file_index)->next.st.sec = sector_buffer[1];
			(pfs->pflex_links+i)->file_id  = NEW_FILE1 - new_file_index;
			(pfs->pflex_links+i)->f_record =
				(pfs->pnew_file+new_file_index)->f_record;
			(pfs->pflex_links+i)->next.st.trk  = sector_buffer[0];
			(pfs->pflex_links+i)->next.st.sec  = sector_buffer[1];
			(pfs->pflex_links+i)->record_nr[0] = sector_buffer[2];
			(pfs->pflex_links+i)->record_nr[1] = sector_buffer[3];
			bytes = fwrite(&sector_buffer[4], 1, 252,
				(pfs->pnew_file+new_file_index)->fp);
			if (bytes != 252) // unable to write sector
				str = 0x10;
			(pfs->pnew_file+new_file_index)->f_record++;
			break;
		default : 
			strcpy((char *)&path, pfs->ppath);
			strcat((char *)&path, "/");
			strcat((char *)&path,
				unix_filename((pfs->pflex_links+i)->file_id));
			(pfs->pflex_links+i)->next.st.trk = sector_buffer[0];
			(pfs->pflex_links+i)->next.st.sec = sector_buffer[1];
			(pfs->pflex_links+i)->record_nr[0]= sector_buffer[2];
			(pfs->pflex_links+i)->record_nr[1]= sector_buffer[3];
			if ((fp = fopen(path, mode)) != NULL &&
				(*mode == 'w' ||
				 !fseek(fp, (pfs->pflex_links+i)->f_record * 252L, SEEK_SET))) {
				bytes = fwrite(&sector_buffer[4], 1, 252, fp);
				fclose(fp);
			} else { // unable to write sector
				str = 0x10;
			} // else
	} // switch
} // write_nafs_sector


int E2floppy::init_nafs_drive(struct s_floppy *pfloppy)
{
	int			i;
	int			wp;	// write protect flag
	struct s_flex_header	header;
	struct stat		sbuf;
	struct s_formats	*pformat;
	char			*mode;
	int			 nafs_number = 0;

	if(stat(pfloppy->ppath, &sbuf))
		return 0;
	if (S_ISDIR(sbuf.st_mode)) {
	  	nafs_number++;
	  	mount_nafs(pfloppy, nafs_number);
	  	return 1;
	} // if
	return 0;
} // init_nafs_drive

void E2floppy::mount_nafs(struct s_floppy *pfloppy, int index)
{
	int wp;

	wp = access(pfloppy->ppath, W_OK);
  	header_for_nafs_format(pfloppy, wp);
	initialize_new_file_table(pfloppy);
  	initialize_flex_sys_info_sectors(pfloppy, index);
  	initialize_flex_unused_sector(pfloppy);
  	fill_flex_directory(pfloppy, wp);
  } // mount_nafs


void E2floppy::umount_nafs(int index)
{
	close_new_files(index);
	free_nafs_memory(index);
}  // umount_nafs

#endif // #ifdef NAFS
